# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark.ml.clustering import KMeans
from pyspark.ml.linalg import Vectors
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from pyspark.ml import Pipeline
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import (
    save_data_models,
    run_onnx_model,
    compare_results,
)
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlKMeansModel(SparkMlTestCase):
    @unittest.skipIf(sys.version_info < (3, 8), reason="pickle fails on python 3.7")
    def test_model_k_means_euclidean(self):
        """
        Testing ONNX conversion for Spark KMeansModel
        when distanceMeasure is set to "euclidean".
        """
        kmeans_euclidean = KMeans(
            k=3,
            distanceMeasure="euclidean",
            featuresCol="features_euclidean",
            predictionCol="prediction_euclidean",
        )
        kmeans_cosine = KMeans(
            k=3,
            distanceMeasure="cosine",
            featuresCol="features_cosine",
            predictionCol="prediction_cosine",
        )

        data = self.spark.createDataFrame(
            [
                (
                    0,
                    Vectors.dense([1.0, 3.1, -1.0]),
                    Vectors.dense([1.0, 1.0, 1.0]),
                ),
                (
                    1,
                    Vectors.dense([1.1, 3.0, -1.1]),
                    Vectors.dense([2.0, 2.0, 2.0]),
                ),
                (
                    2,
                    Vectors.dense([-3.0, 5.1, 9.0]),
                    Vectors.dense([-1.0, 3.0, -5.0]),
                ),
                (
                    3,
                    Vectors.dense([-2.9, 4.9, 8.9]),
                    Vectors.dense([-2.0, 6.0, -10.0]),
                ),
                (
                    4,
                    Vectors.dense([5.0, -3.5, 2.0]),
                    Vectors.dense([1.0, -2.0, 4.0]),
                ),
                (
                    5,
                    Vectors.dense([5.1, -3.3, 2.1]),
                    Vectors.dense([2.0, -4.0, 8.0]),
                ),
            ],
            ["id", "features_euclidean", "features_cosine"],
        )

        model = Pipeline(stages=[kmeans_euclidean, kmeans_cosine]).fit(data)
        model_onnx = convert_sparkml(
            model,
            "Sparkml KMeansModel",
            [
                ("features_euclidean", FloatTensorType([None, 3])),
                ("features_cosine", FloatTensorType([None, 3])),
            ],
            target_opset=TARGET_OPSET,
        )

        self.assertTrue(model_onnx is not None)
        self.assertTrue(model_onnx.graph.node is not None)

        # run the model
        predicted = model.transform(data).toPandas()

        data_pd = data.toPandas()
        data_np = {
            "features_euclidean": data_pd.features_euclidean.apply(
                lambda x: pandas.Series(x.toArray())
            ).values.astype(numpy.float32),
            "features_cosine": data_pd.features_cosine.apply(
                lambda x: pandas.Series(x.toArray())
            ).values.astype(numpy.float32),
        }

        expected = {
            "prediction_euclidean": numpy.asarray(
                predicted.prediction_euclidean.values
            ),
            "prediction_cosine": numpy.asarray(predicted.prediction_cosine.values),
        }

        paths = save_data_models(
            data_np, expected, model, model_onnx, basename="SparkmlKMeansModel"
        )
        onnx_model_path = paths[-1]

        output_names = ["prediction_euclidean", "prediction_cosine"]
        output, output_shapes = run_onnx_model(output_names, data_np, onnx_model_path)
        actual_output = dict(zip(output_names, output))

        assert output_shapes[0] == [None]
        assert output_shapes[1] == [None]
        compare_results(
            expected["prediction_euclidean"],
            actual_output["prediction_euclidean"],
            decimal=5,
        )
        compare_results(
            expected["prediction_cosine"], actual_output["prediction_cosine"], decimal=5
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from ....proto import helper
from ....proto import onnx_proto
from ...common._registration import register_converter
from ...common._apply_operation import apply_constant


def convert_tensor_to_label(scope, operator, container):
    """
    This converter tries to convert a dummy operator 'TensorToLabel' into a sequence of some ONNX operators. Those
    operators are used to extract the label with the highest probability for doing a prediction. We assume that the
    elements in the given probability tensor are aligned with the class labels specified in the CoreML model. That is,
    if you have a class label vector ['a', 'b'] in our CoreML classifier, the first (and the only) input of this
    operator should be [probability_of_class_a, probability_of_class_b].

    Assume that we have C classes with batch size N (N must be 1. If not, the output class probabilities need to be
    encoded as a sequence of dictionary, which is not allowed in ONNX). The ONNX computation graph of this operator may
    look like

                Probability tensor [1, C] (the variable defined at operator.inputs[0])
                        |
                        v
                      ArgMax                                        LoadConstant (its attribute is extracted from
                        |                                                |        operator.raw_operator, which is a
                        |                                                |        CoreML classifier)
                        |                                                |
                        v                                                |
                   best index [1]                                        |
                        |                                                |
                        v                                                v
               ArrayFeatureExtractor  <-------------------- a 1-D tensor of class labels [C]
                        |
                        v
                  predicted label [1]
    """
    model_type = operator.raw_operator.WhichOneof("Type")
    if model_type == "neuralNetworkClassifier":
        model = operator.raw_operator.neuralNetworkClassifier
        if model.WhichOneof("ClassLabels") == "stringClassLabels":
            labels = list(s.encode("utf-8") for s in model.stringClassLabels.vector)
            label_type = onnx_proto.TensorProto.STRING
        elif model.WhichOneof("ClassLabels") == "int64ClassLabels":
            labels = list(int(i) for i in model.int64ClassLabels.vector)
            label_type = onnx_proto.TensorProto.INT64
        else:
            raise ValueError("Unknown label type found")
    elif model_type == "pipelineClassifier":
        model = operator.raw_operator.pipelineClassifier
        if model.WhichOneof("ClassLabels") == "stringClassLabels":
            labels = list(
                s.encode("utf-8")
                for s in model.pipelineClassifier.stringClassLabels.vector
            )
            label_type = onnx_proto.TensorProto.STRING
        elif model.WhichOneof("ClassLabels") == "int64ClassLabels":
            labels = list(int(i) for i in model.int64ClassLabels.vector)
            label_type = onnx_proto.TensorProto.INT64
        else:
            raise ValueError("Unknown label type found")
    else:
        raise ValueError(
            "Only neural network classifiers and pipeline classifiers are supported"
        )

    # Use a Constant operator to load and output all labels as a tensor
    label_loader_name = scope.get_unique_operator_name("LabelLoader")
    label_buffer_name = scope.get_unique_variable_name("ClassLabels")
    label_loader_value = helper.make_tensor(
        label_buffer_name, label_type, [len(labels)], labels
    )
    apply_constant(
        scope,
        [label_buffer_name],
        container,
        operator_name=label_loader_name,
        value=label_loader_value,
    )

    # Extract most possible label index
    label_id_extractor_name = scope.get_unique_operator_name("LabelIndexExtractor")
    label_id_extractor_attrs = {"name": label_id_extractor_name}
    label_id_extractor_attrs["axis"] = 1
    label_id_extractor_attrs["keepdims"] = 1
    extracted_id_name = scope.get_unique_variable_name("LabelId")
    container.add_node(
        "ArgMax",
        [operator.inputs[0].full_name],
        [extracted_id_name],
        **label_id_extractor_attrs
    )

    # Pick up the label indicated by the selected ID
    label_selector_name = scope.get_unique_operator_name("LabelSelector")
    label_selector_attrs = {"name": label_selector_name}
    container.add_node(
        "ArrayFeatureExtractor",
        [label_buffer_name, extracted_id_name],
        [operator.outputs[0].full_name],
        op_domain="ai.onnx.ml",
        **label_selector_attrs
    )


register_converter("tensorToLabel", convert_tensor_to_label)



# SPDX-License-Identifier: Apache-2.0
import logging
from ...common.data_types import FloatTensorType
from ...common.tree_ensemble import (
    add_tree_to_attribute_pairs,
    get_default_tree_regressor_attribute_pairs,
)
from ...common.utils import check_input_and_output_numbers
from ...sparkml.operator_converters.decision_tree_classifier import (
    save_read_sparkml_model_data,
)
from ...sparkml.operator_converters.tree_ensemble_common import (
    sparkml_tree_dataset_to_sklearn,
)
from ...common._registration import register_converter, register_shape_calculator
from .tree_helper import rewrite_ids_and_process

logger = logging.getLogger("onnxmltools")


def convert_decision_tree_regressor(scope, operator, container):
    op = operator.raw_operator
    op_type = "TreeEnsembleRegressor"

    attrs = get_default_tree_regressor_attribute_pairs()
    attrs["name"] = scope.get_unique_operator_name(op_type)
    attrs["n_targets"] = 1

    tree_df = save_read_sparkml_model_data(operator.raw_params["SparkSession"], op)
    tree = sparkml_tree_dataset_to_sklearn(tree_df, is_classifier=False)
    add_tree_to_attribute_pairs(attrs, False, tree, 0, 1.0, 0, False)
    new_attrs = rewrite_ids_and_process(attrs, logger)

    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **new_attrs
    )


register_converter(
    "pyspark.ml.regression.DecisionTreeRegressionModel", convert_decision_tree_regressor
)


def calculate_decision_tree_regressor_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType(shape=[N, 1])


register_shape_calculator(
    "pyspark.ml.regression.DecisionTreeRegressionModel",
    calculate_decision_tree_regressor_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from .....proto import onnx_proto
from ....common._registration import register_converter
from .Reshape import apply_reshape
from .SimpleRNN import extract_rnn_activation_info


def convert_unidirectional_lstm(scope, operator, container):
    # The LSTM inputs are feature vector, X, initial hidden state,
    # h_init, and initial cell state, c_init.
    # In CorML, their shapes respectively are [S, C_in], [1, C_out],
    # and [1, C_out], where C_in is input feature
    # length, # C_out is output dimension, and S is sequence length.
    # Note that S-axis is also known as time axis.
    # In ONNX, those shapes become [S, N, C_in] (X), [D, N, C_out]
    # (h_init), and [D, N, C_out]. To simulate
    # CoreML LSTM under ONNX, we need some extra operators
    # in addition to LSTM itself.
    #
    # Note that N=1 and D=1 are always true in ONNX
    # if we are considering LSTM in CoreML because there is no
    # batch size in CoreML spec and CoreML LSTM
    # is always uni-directional.
    #
    # Below we provide a visualization of our conversion for CoreML LSTM.
    #
    # Symbols:
    #
    #  X: input features of CoreML LSTM
    #  h_init: initial LSTM hidden state in CoreML
    #  c_init: initial LSTM cell state in CoreML
    #  Y: CoreML LSTM's output. It can be [S, C_out]
    # (if sequence_output is on) or [1, C_out] (if sequence_output is off)
    #  Y_h: CoreML LSTM's last hidden state
    #  Y_c: CoreML LSTM's last cell state
    #
    #  X': input features of ONNX LSTM
    #  h_init': initial LSTM hidden state of ONNX
    #  c_init': initial LSTM hidden state of ONNX
    #  Y': ONNX LSTM's output
    #  Y_h': ONNX LSTM's last hidden state
    #  Y_c': ONNX LSTM's last cell state
    #
    # Computational graph of CoreML LSTM (if sequence_output is on):
    #
    #      X [S, C_in]     h_init [1, C_out]    c_init [1, C_out]
    #      |                 |                    |
    #      v                 v                    v
    #      '---------------  |    -----------------
    #                     |  |    |
    #                     v  v    v
    #                    CoreML LSTM
    #                     |  |    |
    #      ---------------'  l    '----------------
    #      |                 |                    |
    #      v                 v                    v
    #      Y [S, C_out]     Y_h [1, C_out]       Y_c [1, C_out]
    #
    # Computational graph of CoreML LSTM in ONNX (if sequence_output is on):
    #
    #      X [S, C_in]     h_init [1, C_out]    c_init [1, C_out]
    #      |                 |                    |
    #      v                 v                    v
    #   Reshape           Reshape              Reshape
    #      |                 |                    |
    #      v                 v                    v
    #      X'[S, 1, C_in]  h_init'[1, 1, C_out] c_init [1, 1, C_out]
    #      |                 |                    |
    #      '---------------  |    -----------------
    #                     |  |    |
    #                     v  v    v
    #                     ONNX LSTM
    #                     |  |    |
    #       --------------'  vl   '---------------
    #      |                 |                    |
    #      v                 v                    v
    #      Y'[S, 1, C_out]  Y_h' [1, 1, C_out]   Y_c [1, 1, C_out]
    #      |                 |                    |
    #      v                 v                    v
    #   Reshape           Reshape              Reshape
    #      |                 |                    |
    #      v                 v                    v
    #      Y [S, C_out]     Y_h [1, C_out]       Y_c [1, C_out]
    #
    # Computational graph of CoreML LSTM (if sequence_output is off):
    #
    #      X [S, C_in]     h_init [1, C_out]    c_init [1, C_out]
    #      |                 |                    |
    #      v                 v                    v
    #      '---------------  |    -----------------
    #                     |  |    |
    #                     v  v    v
    #                    CoreML LSTM
    #                     |  |    |
    #      ---------------'  l    '----------------
    #      |                 |                    |
    #      v                 v                    v
    #      Y [1, C_out]     Y_h [1, C_out]       Y_c [1, C_out]
    #           (Note that Y = Y_h)
    #
    # Computational graph of CoreML LSTM in ONNX (if sequence_output is off):
    #
    #      X [S, C_in]     h_init [1, C_out]    c_init [1, C_out]
    #      |                 |                    |
    #      v                 v                    v
    #   Reshape           Reshape              Reshape
    #      |                 |                    |
    #      v                 v                    v
    #      X'[S, 1, C_in]  h_init'[1, 1, C_out] c_init [1, 1, C_out]
    #      |                 |                    |
    #      '---------------  |    -----------------
    #                     |  |    |
    #                     v  v    v
    #                     ONNX LSTM
    #                     |  |    |
    #       --------------'  vl   '---------------
    #      |                 |                    |
    #      v                 v                    v
    #      Y'[S, 1, C_out]  Y_h' [1, 1, C_out]   Y_c [1, 1, C_out]
    #  (useless output)      |                    |
    #                        v                    v
    #                     Reshape              Reshape
    #                        |                    |
    #                        v                    v
    #                    Y [1, C_out]       Y_c [1, C_out]
    #                        |
    #                        v
    #                     Identity
    #                        |
    #                        v
    #                    Y_h [1, C_out]

    params = operator.raw_operator.uniDirectionalLSTM
    lstm_params = params.params
    lstm_weights = params.weightParams
    input_size = params.inputVectorSize
    hidden_size = params.outputVectorSize

    # Initialize materials needed to create ONNX LSTM
    lstm_op_name = scope.get_unique_operator_name("LSTM")
    lstm_attrs = {"name": lstm_op_name}
    lstm_inputs = []
    lstm_outputs = []

    # Reshape input feature vector in CoreML format into ONNX format
    lstm_x_reshape_name = scope.get_unique_variable_name(lstm_op_name + "_X_reshape")
    apply_reshape(
        scope,
        operator.inputs[0].full_name,
        lstm_x_reshape_name,
        container,
        desired_shape=[-1, 1, input_size],
    )
    lstm_inputs.append(lstm_x_reshape_name)

    # Allocate LSTM's weight matrices and add them into ONNX LSTM's input list
    matrices_w = np.concatenate(
        [
            lstm_weights.inputGateWeightMatrix.floatValue,
            lstm_weights.outputGateWeightMatrix.floatValue,
            lstm_weights.forgetGateWeightMatrix.floatValue,
            lstm_weights.blockInputWeightMatrix.floatValue,
        ]
    )
    matrices_w_name = scope.get_unique_variable_name(lstm_op_name + "_W")
    container.add_initializer(
        matrices_w_name,
        onnx_proto.TensorProto.FLOAT,
        [1, 4 * hidden_size, input_size],
        matrices_w,
    )
    lstm_inputs.append(matrices_w_name)

    # Allocate LSTM's recursion weight matrices and add them into ONNX LSTM's input list
    matrices_r = np.concatenate(
        [
            lstm_weights.inputGateRecursionMatrix.floatValue,
            lstm_weights.outputGateRecursionMatrix.floatValue,
            lstm_weights.forgetGateRecursionMatrix.floatValue,
            lstm_weights.blockInputRecursionMatrix.floatValue,
        ]
    )
    matrices_r_name = scope.get_unique_variable_name(lstm_op_name + "_R")
    container.add_initializer(
        matrices_r_name,
        onnx_proto.TensorProto.FLOAT,
        [1, 4 * hidden_size, hidden_size],
        matrices_r,
    )
    lstm_inputs.append(matrices_r_name)

    # Handle bias vectors
    vectors_b = np.zeros(shape=(8, hidden_size))
    if lstm_params.hasBiasVectors:
        vectors_b[0, :] = lstm_weights.inputGateBiasVector.floatValue
        vectors_b[1, :] = lstm_weights.outputGateBiasVector.floatValue
        vectors_b[2, :] = lstm_weights.forgetGateBiasVector.floatValue
        vectors_b[3, :] = lstm_weights.blockInputBiasVector.floatValue
    if lstm_params.forgetBias:
        # One may think we should do something like b[2, :] += 1., but it's wrong as CoreML has
        # added 1 into lstm_weights.forgetGateBiasVector.floatValue.
        pass
    if lstm_params.hasBiasVectors or lstm_params.forgetBias:
        vectors_b_name = scope.get_unique_variable_name(lstm_op_name + "_B")
        container.add_initializer(
            vectors_b_name,
            onnx_proto.TensorProto.FLOAT,
            [1, 8 * hidden_size],
            vectors_b.flatten(),
        )
        lstm_inputs.append(vectors_b_name)
    else:
        lstm_inputs.append("")

    # Converting CoreML LSTM doesn't need sequence length
    lstm_inputs.append("")

    # Provide ONNX LSTM the initial hidden state when necessary
    if len(operator.inputs) > 1:
        # Assign a Reshape to adjust CoreML hidden state's shape
        # [1, C]/[1, C, 1, 1] into its ONNX counterpart [1, 1, C]
        lstm_h_init_reshape_name = scope.get_unique_variable_name(
            lstm_op_name + "_h_init_reshape"
        )
        apply_reshape(
            scope,
            operator.inputs[1].full_name,
            lstm_h_init_reshape_name,
            container,
            desired_shape=[1, 1, hidden_size],
        )
        lstm_inputs.append(lstm_h_init_reshape_name)

        # Add a zero initializer to initial hidden state so that this variable becomes optional
        container.add_initializer(
            operator.inputs[1].full_name,
            onnx_proto.TensorProto.FLOAT,
            operator.inputs[1].type.shape,
            np.zeros(shape=operator.inputs[1].type.shape).flatten(),
        )
    else:
        lstm_inputs.append("")

    # Provide ONNX LSTM the initial cell state when necessary
    if len(operator.inputs) > 2:
        lstm_c_init_reshape_name = scope.get_unique_variable_name(
            lstm_op_name + "_c_init_reshape"
        )
        apply_reshape(
            scope,
            operator.inputs[2].full_name,
            lstm_c_init_reshape_name,
            container,
            desired_shape=[1, 1, hidden_size],
        )
        lstm_inputs.append(lstm_c_init_reshape_name)

        # Add a zero initializer to initial cell state so that this variable becomes optional
        container.add_initializer(
            operator.inputs[2].full_name,
            onnx_proto.TensorProto.FLOAT,
            operator.inputs[2].type.shape,
            np.zeros(shape=operator.inputs[2].type.shape).flatten(),
        )
    else:
        lstm_inputs.append("")

    # Add peephole vector when presenting
    if lstm_params.hasPeepholeVectors:
        vectors_p = np.concatenate(
            [
                lstm_weights.inputGatePeepholeVector.floatValue,
                lstm_weights.outputGatePeepholeVector.floatValue,
                lstm_weights.forgetGatePeepholeVector.floatValue,
            ]
        )
        vectors_p_name = scope.get_unique_variable_name(lstm_op_name + "_P")
        container.add_initializer(
            vectors_p_name,
            onnx_proto.TensorProto.FLOAT,
            [1, 3 * hidden_size],
            vectors_p,
        )
        lstm_inputs.append(vectors_p_name)
    else:
        lstm_inputs.append("")

    # Parse activation functions' information and add
    # them into ONNX LSTM's attribute dictionary
    activation_types = []
    alphas = []
    betas = []
    for activation in params.activations:
        activation_type, alpha, beta = extract_rnn_activation_info(activation)
        activation_types.append(activation_type.encode("utf-8"))
        if alpha is not None:
            alphas.append(alpha)
        if beta is not None:
            betas.append(beta)
    lstm_attrs["activations"] = activation_types
    if alphas:
        lstm_attrs["activation_alpha"] = alphas
    if betas:
        lstm_attrs["activation_beta"] = betas

    # Set up other attributes
    lstm_attrs["direction"] = "reverse" if params.reverseInput else "forward"
    lstm_attrs["hidden_size"] = hidden_size
    lstm_attrs["clip"] = float(lstm_params.cellClipThreshold)
    lstm_attrs["input_forget"] = lstm_params.coupledInputAndForgetGate

    # Set up version-dependent attributes
    if container.target_opset < 7:
        lstm_attrs["output_sequence"] = lstm_params.sequenceOutput
        op_version = 1
    else:
        op_version = 7

    # Create the main LSTM operator
    lstm_y_name = scope.get_unique_variable_name(lstm_op_name + "_Y")
    lstm_y_h_name = scope.get_unique_variable_name(lstm_op_name + "_Y_h")
    lstm_c_name = scope.get_unique_variable_name(lstm_op_name + "_Y_c")
    lstm_outputs.extend([lstm_y_name, lstm_y_h_name, lstm_c_name])
    container.add_node(
        "LSTM", lstm_inputs, lstm_outputs, op_version=op_version, **lstm_attrs
    )

    # Handle the first output of LSTM
    if lstm_params.sequenceOutput:
        # Handle the first output of LSTM
        apply_reshape(
            scope,
            lstm_y_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[-1, hidden_size],
        )

        # Handle the second output of LSTM
        if len(operator.outputs) > 1:
            apply_reshape(
                scope,
                lstm_y_h_name,
                operator.outputs[1].full_name,
                container,
                desired_shape=[1, hidden_size],
            )
    else:
        # Here we ingore ONNX LSTM's first output because
        # it's useless and use the second output of ONNX LSTM to produce
        # the first output of CoreML LSTM
        apply_reshape(
            scope,
            lstm_y_h_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[1, hidden_size],
        )

        # Create the second LSTM output from the first output
        if len(operator.outputs) > 1:
            container.add_node(
                "Identity",
                operator.outputs[0].full_name,
                operator.outputs[1].full_name,
                name=scope.get_unique_operator_name("Identity"),
            )

    # Handle the cell state output of LSTM
    if len(operator.outputs) > 2:
        apply_reshape(
            scope,
            lstm_c_name,
            operator.outputs[2].full_name,
            container,
            desired_shape=[1, hidden_size],
        )


register_converter("uniDirectionalLSTM", convert_unidirectional_lstm)



# SPDX-License-Identifier: Apache-2.0

from uuid import uuid4
import onnx
from ..common.onnx_ex import get_maximum_opset_supported
from ..common._topology import convert_topology
from ._parse import parse_libsvm

# Invoke the registration of all our converters and shape calculators


def convert(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=onnx.__version__,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    """
    :param model: a libsvm model
    :param initial_types: a python list. Each element is a
        tuple of a variable name and a type defined in data_types.py
    :param name: The name of the graph (type: GraphProto)
        in the produced ONNX model (type: ModelProto)
    :param doc_string: A string attached onto the produced ONNX model
    :param target_opset: number, for example, 7 for
        ONNX 1.2, and 8 for ONNX 1.3.
    :param targeted_onnx: A string (for example, '1.1.2' and '1.2')
        used to specify the targeted ONNX version of the
        produced model. If ONNXMLTools cannot find a compatible
        ONNX python package, an error may be thrown.
    :param custom_conversion_functions: a dictionary for
        specifying the user customized conversion function
    :param custom_shape_calculators: a dictionary for
        specifying the user customized shape calculator
    :return: An ONNX model (type: ModelProto) which is
        equivalent to the input scikit-learn model
    """
    if initial_types is None:
        raise ValueError(
            "Initial types are required. See usage of convert(...) in \
                         onnxmltools.convert.libsvm.convert for details"
        )

    if name is None:
        name = str(uuid4().hex)
    target_opset = target_opset if target_opset else get_maximum_opset_supported()

    # Parse scikit-learn model as our internal data structure (i.e., Topology)
    topology = parse_libsvm(
        model, initial_types, custom_conversion_functions, custom_shape_calculators
    )

    # Infer variable shapes
    topology.compile()

    # Convert our Topology object into ONNX. The outcome is an ONNX model.
    onnx_model = convert_topology(
        topology, name, doc_string, target_opset, targeted_onnx
    )

    return onnx_model



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_reorganize_data_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N, C * B * B , H / B, W / B]
        2. [N, C, H, W] ---> [N, C / B / B , H * B, W * B]

    Note taht B is the block size specified in this operator.
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)

    params = operator.raw_operator.reorganizeData

    from coremltools.proto.NeuralNetwork_pb2 import ReorganizeDataLayerParams as Params

    if params.mode == Params.DEPTH_TO_SPACE:
        if output_shape[1] % (params.blockSize * params.blockSize) != 0:
            raise RuntimeError(
                "Channel number must be divisible by the square of block size"
            )

        output_shape = [
            output_shape[0],
            output_shape[1] / params.blockSize / params.blockSize,
            output_shape[2] * params.blockSize,
            output_shape[3] * params.blockSize,
        ]
    elif params.mode == Params.SPACE_TO_DEPTH:
        if (
            output_shape[2] % params.blockSize != 0
            or output_shape[3] % params.blockSize != 0
        ):
            raise RuntimeError("Height and weight must be divisible by block size")

        output_shape = [
            output_shape[0],
            output_shape[1] * params.blockSize * params.blockSize,
            output_shape[2] / params.blockSize,
            output_shape[3] / params.blockSize,
        ]
    else:
        raise ValueError("Unsupport reorganization mode {0}".format(params.mode))

    operator.outputs[0].type = FloatTensorType(
        [int(i) if i != "None" else "None" for i in output_shape],
        doc_string=operator.outputs[0].type.doc_string,
    )


register_shape_calculator("reorganizeData", calculate_reorganize_data_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_upsample_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N, C, H', W']
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    scales = operator.raw_operator.upsample.scalingFactor

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)
    output_shape[2] *= scales[0]
    output_shape[3] *= scales[1]

    operator.outputs[0].type = FloatTensorType(
        output_shape, doc_string=operator.outputs[0].type.doc_string
    )


register_shape_calculator("upsample", calculate_upsample_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_pad
from ....common._registration import register_converter


def convert_padding(scope, operator, container):
    params = operator.raw_operator.padding

    pad_table = {"constant": "constant", "reflection": "reflect", "replication": "edge"}
    pad_type = params.WhichOneof("PaddingType")
    if pad_type not in pad_table:
        raise ValueError("Unsupported padding mode: {}".format(pad_type))
    mode = pad_table[pad_type]

    # CoreML only pads for their H- and W-axes.
    # Here we assume the shape of the tensor to be padded
    # is [N, C, H, W], so we have 8 padding amounts
    #     pads = [N_begin_index, C_begin_index,
    #             H_begin_index, W_begin_index,
    #             N_end_index,   C_end_index,
    #             H_end_index,   W_end_index]
    # Because only H- and W-axes are padded in CoreML,
    # we leave padding amounts of N- and C-axes zeros.
    pads = [0, 0, 0, 0, 0, 0, 0, 0]
    if len(params.paddingAmounts.borderAmounts) > 0:
        # Set H_begin_index
        pads[2] = params.paddingAmounts.borderAmounts[0].startEdgeSize
        # Set W_begin_index
        pads[3] = params.paddingAmounts.borderAmounts[1].startEdgeSize
        # Set H_end_index
        pads[6] = params.paddingAmounts.borderAmounts[0].endEdgeSize
        # Set W_end_index
        pads[7] = params.paddingAmounts.borderAmounts[1].endEdgeSize

    if pad_type == "constant":
        value = params.constant.value
    else:
        value = None

    apply_pad(
        scope,
        operator.input_full_names,
        operator.output_full_names,
        container,
        operator_name=operator.full_name,
        mode=mode,
        pads=pads,
        value=value,
    )


register_converter("padding", convert_padding)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_lrn(scope, operator, container):
    op_type = "LRN"
    params = operator.raw_operator.lrn
    attrs = {"name": scope.get_unique_operator_name(op_type)}
    attrs["size"] = params.localSize
    attrs["alpha"] = float(params.alpha)
    attrs["beta"] = float(params.beta)
    attrs["bias"] = float(params.k)
    container.add_node(
        op_type, operator.input_full_names, operator.output_full_names, **attrs
    )


register_converter("lrn", convert_lrn)



# SPDX-License-Identifier: Apache-2.0

import numpy
from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_floor, apply_div, apply_matmul
from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from .tree_ensemble_common import save_read_sparkml_model_data

g_rand_vectors = None


def get_rand_vectors(operator):
    global g_rand_vectors
    if not g_rand_vectors:
        g_rand_vectors = (
            save_read_sparkml_model_data(
                operator.raw_params["SparkSession"], operator.raw_operator
            )
            .first()[0]
            .toArray()
            .transpose()
            .astype(numpy.float32)
        )
    return g_rand_vectors


def convert_min_hash_lsh(scope, operator, container):
    rand_vectors = get_rand_vectors(operator)
    bucket_length = float(operator.raw_operator.getOrDefault("bucketLength"))

    rand_vectors_tensor = scope.get_unique_variable_name("rand_vectors_tensor")
    container.add_initializer(
        rand_vectors_tensor,
        onnx_proto.TensorProto.FLOAT,
        rand_vectors.shape,
        rand_vectors.flatten(),
    )
    matmul_result = scope.get_unique_variable_name("matmul_result_tensor")
    apply_matmul(
        scope,
        [operator.input_full_names[0], rand_vectors_tensor],
        matmul_result,
        container,
    )
    bucket_length_tensor = scope.get_unique_variable_name("bucket_length_tensor")
    container.add_initializer(
        bucket_length_tensor, onnx_proto.TensorProto.FLOAT, [1], [bucket_length]
    )
    div_result = scope.get_unique_variable_name("div_result_tensor")
    apply_div(scope, [matmul_result, bucket_length_tensor], div_result, container)
    apply_floor(scope, div_result, operator.output_full_names[0], container)


register_converter(
    "pyspark.ml.feature.BucketedRandomProjectionLSHModel", convert_min_hash_lsh
)


def calculate_min_hash_lsh_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    N = operator.inputs[0].type.shape[0]
    C = 1
    operator.outputs[0].type = FloatTensorType([N, C])


register_shape_calculator(
    "pyspark.ml.feature.BucketedRandomProjectionLSHModel",
    calculate_min_hash_lsh_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import copy
import numbers
from collections import deque, Counter
import ctypes
import json
import numpy as np
from onnx import TensorProto
from ...common._apply_operation import (
    apply_div,
    apply_reshape,
    apply_sub,
    apply_cast,
    apply_identity,
)
from ...common._registration import register_converter
from ...common.tree_ensemble import get_default_tree_classifier_attribute_pairs
from ....proto import onnx_proto


def has_tqdm():
    try:
        from tqdm import tqdm  # noqa

        return True
    except ImportError:
        return False


def _translate_split_criterion(criterion):
    # If the criterion is true, LightGBM use the left child.
    # Otherwise, right child is selected.
    if criterion == "<=":
        return "BRANCH_LEQ"
    elif criterion == "<":
        return "BRANCH_LT"
    elif criterion == ">=":
        return "BRANCH_GTE"
    elif criterion == ">":
        return "BRANCH_GT"
    elif criterion == "==":
        return "BRANCH_EQ"
    elif criterion == "!=":
        return "BRANCH_NEQ"
    else:
        raise ValueError(
            "Unsupported splitting criterion: %s. Only <=, " "<, >=, and > are allowed."
        )


def _create_node_id(node_id_pool):
    i = 0
    while i in node_id_pool:
        i += 1
    node_id_pool.add(i)
    return i


def _parse_tree_structure(tree_id, class_id, learning_rate, tree_structure, attrs):
    """
    The pool of all nodes' indexes created when parsing a single tree.
    Different tree use different pools.
    """
    node_id_pool = set()
    node_pyid_pool = dict()

    node_id = _create_node_id(node_id_pool)
    node_pyid_pool[id(tree_structure)] = node_id

    # The root node is a leaf node.
    if "left_child" not in tree_structure or "right_child" not in tree_structure:
        _parse_node(
            tree_id,
            class_id,
            node_id,
            node_id_pool,
            node_pyid_pool,
            learning_rate,
            tree_structure,
            attrs,
        )
        return

    left_pyid = id(tree_structure["left_child"])
    right_pyid = id(tree_structure["right_child"])

    if left_pyid in node_pyid_pool:
        left_id = node_pyid_pool[left_pyid]
        left_parse = False
    else:
        left_id = _create_node_id(node_id_pool)
        node_pyid_pool[left_pyid] = left_id
        left_parse = True

    if right_pyid in node_pyid_pool:
        right_id = node_pyid_pool[right_pyid]
        right_parse = False
    else:
        right_id = _create_node_id(node_id_pool)
        node_pyid_pool[right_pyid] = right_id
        right_parse = True

    attrs["nodes_treeids"].append(tree_id)
    attrs["nodes_nodeids"].append(node_id)

    attrs["nodes_featureids"].append(tree_structure["split_feature"])
    attrs["nodes_modes"].append(
        _translate_split_criterion(tree_structure["decision_type"])
    )
    if isinstance(tree_structure["threshold"], str):
        try:
            attrs["nodes_values"].append(float(tree_structure["threshold"]))
        except ValueError:
            import pprint

            text = pprint.pformat(tree_structure)
            if len(text) > 100000:
                text = text[:100000] + "\n..."
            raise TypeError(
                "threshold must be a number not '{}'"
                "\n{}".format(tree_structure["threshold"], text)
            )
    else:
        attrs["nodes_values"].append(tree_structure["threshold"])

    # Assume left is the true branch and right is the false branch
    attrs["nodes_truenodeids"].append(left_id)
    attrs["nodes_falsenodeids"].append(right_id)
    if tree_structure["default_left"]:
        if (
            tree_structure["missing_type"] == "None"
            and float(tree_structure["threshold"]) < 0.0
        ):
            attrs["nodes_missing_value_tracks_true"].append(0)
        else:
            attrs["nodes_missing_value_tracks_true"].append(1)
    else:
        attrs["nodes_missing_value_tracks_true"].append(0)
    attrs["nodes_hitrates"].append(1.0)
    if left_parse:
        _parse_node(
            tree_id,
            class_id,
            left_id,
            node_id_pool,
            node_pyid_pool,
            learning_rate,
            tree_structure["left_child"],
            attrs,
        )
    if right_parse:
        _parse_node(
            tree_id,
            class_id,
            right_id,
            node_id_pool,
            node_pyid_pool,
            learning_rate,
            tree_structure["right_child"],
            attrs,
        )


def _parse_node(
    tree_id, class_id, node_id, node_id_pool, node_pyid_pool, learning_rate, node, attrs
):
    """
    Parses nodes.
    """
    if (hasattr(node, "left_child") and hasattr(node, "right_child")) or (
        "left_child" in node and "right_child" in node
    ):
        left_pyid = id(node["left_child"])
        right_pyid = id(node["right_child"])

        if left_pyid in node_pyid_pool:
            left_id = node_pyid_pool[left_pyid]
            left_parse = False
        else:
            left_id = _create_node_id(node_id_pool)
            node_pyid_pool[left_pyid] = left_id
            left_parse = True

        if right_pyid in node_pyid_pool:
            right_id = node_pyid_pool[right_pyid]
            right_parse = False
        else:
            right_id = _create_node_id(node_id_pool)
            node_pyid_pool[right_pyid] = right_id
            right_parse = True

        attrs["nodes_treeids"].append(tree_id)
        attrs["nodes_nodeids"].append(node_id)

        attrs["nodes_featureids"].append(node["split_feature"])
        attrs["nodes_modes"].append(_translate_split_criterion(node["decision_type"]))
        if isinstance(node["threshold"], str):
            try:
                attrs["nodes_values"].append(float(node["threshold"]))
            except ValueError:
                import pprint

                text = pprint.pformat(node)
                if len(text) > 100000:
                    text = text[:100000] + "\n..."
                raise TypeError(
                    "threshold must be a number not '{}'"
                    "\n{}".format(node["threshold"], text)
                )
        else:
            attrs["nodes_values"].append(node["threshold"])

        # Assume left is the true branch
        # and right is the false branch
        attrs["nodes_truenodeids"].append(left_id)
        attrs["nodes_falsenodeids"].append(right_id)
        if node["default_left"]:
            if node["missing_type"] == "None" and float(node["threshold"]) < 0.0:
                attrs["nodes_missing_value_tracks_true"].append(0)
            else:
                attrs["nodes_missing_value_tracks_true"].append(1)
        else:
            attrs["nodes_missing_value_tracks_true"].append(0)
        attrs["nodes_hitrates"].append(1.0)

        # Recursively dive into the child nodes
        if left_parse:
            _parse_node(
                tree_id,
                class_id,
                left_id,
                node_id_pool,
                node_pyid_pool,
                learning_rate,
                node["left_child"],
                attrs,
            )
        if right_parse:
            _parse_node(
                tree_id,
                class_id,
                right_id,
                node_id_pool,
                node_pyid_pool,
                learning_rate,
                node["right_child"],
                attrs,
            )
    elif hasattr(node, "left_child") or hasattr(node, "right_child"):
        raise ValueError("Need two branches")
    else:
        # Node attributes
        attrs["nodes_treeids"].append(tree_id)
        attrs["nodes_nodeids"].append(node_id)
        attrs["nodes_featureids"].append(0)
        attrs["nodes_modes"].append("LEAF")
        # Leaf node has no threshold.
        # A zero is appended but it will never be used.
        attrs["nodes_values"].append(0.0)
        # Leaf node has no child.
        # A zero is appended but it will never be used.
        attrs["nodes_truenodeids"].append(0)
        # Leaf node has no child.
        # A zero is appended but it will never be used.
        attrs["nodes_falsenodeids"].append(0)
        # Leaf node has no split function.
        # A zero is appended but it will never be used.
        attrs["nodes_missing_value_tracks_true"].append(0)
        attrs["nodes_hitrates"].append(1.0)

        # Leaf attributes
        attrs["class_treeids"].append(tree_id)
        attrs["class_nodeids"].append(node_id)
        attrs["class_ids"].append(class_id)
        attrs["class_weights"].append(float(node["leaf_value"]) * learning_rate)


def dump_booster_model(
    self, num_iteration=None, start_iteration=0, importance_type="split", verbose=0
):
    """
    Dumps Booster to JSON format.

    Parameters
    ----------
    self: booster
    num_iteration : int or None, optional (default=None)
        Index of the iteration that should be dumped.
        If None, if the best iteration exists, it is dumped; otherwise,
        all iterations are dumped.
        If <= 0, all iterations are dumped.
    start_iteration : int, optional (default=0)
        Start index of the iteration that should be dumped.
    importance_type : string, optional (default="split")
        What type of feature importance should be dumped.
        If "split", result contains numbers of times the feature is used in a model.
        If "gain", result contains total gains of splits which use the feature.
    verbose: dispays progress (usefull for big trees)

    Returns
    -------
    json_repr : dict
        JSON format of Booster.

    .. note::
        This function is inspired from
        the *lightgbm* (`dump_model
        <https://lightgbm.readthedocs.io/en/latest/pythonapi/
        lightgbm.Booster.html#lightgbm.Booster.dump_model>`_.
        It creates intermediate structure to speed up the conversion
        into ONNX of such model. The function overwrites the
        `json.load` to fastly extract nodes.
    """
    if getattr(self, "is_mock", False):
        return self.dump_model(), None
    from lightgbm.basic import _LIB, _safe_call

    try:
        # lightgbm >= 4.0
        from lightgbm.basic import (
            _FEATURE_IMPORTANCE_TYPE_MAPPER as FITM,
            _json_default_with_numpy as jdwn,
        )
    except ImportError:
        # lightgbm < 4.0
        from lightgbm.basic import (
            FEATURE_IMPORTANCE_TYPE_MAPPER as FITM,
            json_default_with_numpy as jdwn,
        )
    if num_iteration is None:
        num_iteration = self.best_iteration
    importance_type_int = FITM[importance_type]
    buffer_len = 1 << 20
    tmp_out_len = ctypes.c_int64(0)
    string_buffer = ctypes.create_string_buffer(buffer_len)
    ptr_string_buffer = ctypes.c_char_p(*[ctypes.addressof(string_buffer)])
    if verbose >= 2:
        print("[dump_booster_model] call CAPI: LGBM_BoosterDumpModel")
    try:
        handle = self._handle
    except AttributeError:
        handle = self.handle
    _safe_call(
        _LIB.LGBM_BoosterDumpModel(
            handle,
            ctypes.c_int(start_iteration),
            ctypes.c_int(num_iteration),
            ctypes.c_int(importance_type_int),
            ctypes.c_int64(buffer_len),
            ctypes.byref(tmp_out_len),
            ptr_string_buffer,
        )
    )
    actual_len = tmp_out_len.value
    # if buffer length is not long enough, reallocate a buffer
    if actual_len > buffer_len:
        string_buffer = ctypes.create_string_buffer(actual_len)
        ptr_string_buffer = ctypes.c_char_p(*[ctypes.addressof(string_buffer)])
        try:
            # lightgbm >= 4.0
            handle = self._handle
        except AttributeError:
            # lightgbm < 4.0
            handle = self.handle
        _safe_call(
            _LIB.LGBM_BoosterDumpModel(
                handle,
                ctypes.c_int(start_iteration),
                ctypes.c_int(num_iteration),
                ctypes.c_int(importance_type_int),
                ctypes.c_int64(buffer_len),
                ctypes.byref(tmp_out_len),
                ptr_string_buffer,
            )
        )
    actual_len = tmp_out_len.value
    # if buffer length is not long enough, reallocate a buffer
    if actual_len > buffer_len:
        string_buffer = ctypes.create_string_buffer(actual_len)
        ptr_string_buffer = ctypes.c_char_p(*[ctypes.addressof(string_buffer)])
        try:
            handle = self._handle
        except AttributeError:
            handle = self.handle
        _safe_call(
            _LIB.LGBM_BoosterDumpModel(
                handle,
                ctypes.c_int(start_iteration),
                ctypes.c_int(num_iteration),
                ctypes.c_int(importance_type_int),
                ctypes.c_int64(actual_len),
                ctypes.byref(tmp_out_len),
                ptr_string_buffer,
            )
        )

    class Hook(json.JSONDecoder):
        """
        Keep track of the progress, stores a copy of all objects with
        a decision into a different container in order to walk through
        all nodes in a much faster way than going through the architecture.
        """

        def __init__(self, *args, info=None, n_trees=None, verbose=0, **kwargs):
            json.JSONDecoder.__init__(self, object_hook=self.hook, *args, **kwargs)
            self.nodes = []
            self.buffer = []
            self.info = info
            self.n_trees = n_trees
            self.verbose = verbose
            self.stored = 0
            if verbose >= 2 and n_trees is not None and has_tqdm():
                from tqdm import tqdm

                self.loop = tqdm(total=n_trees)
                self.loop.set_description("dump_booster")
            else:
                self.loop = None

        def hook(self, obj):
            """
            Hook called everytime a JSON object is created.
            Keep track of the progress, stores a copy of all objects with
            a decision into a different container.
            """
            # Every obj goes through this function from the leaves to the root.
            if "tree_info" in obj:
                self.info["decision_nodes"] = self.nodes
                if self.n_trees is not None and len(self.nodes) != self.n_trees:
                    raise RuntimeError(
                        "Unexpected number of trees %d (expecting %d)."
                        % (len(self.nodes), self.n_trees)
                    )
                self.nodes = []
                if self.loop is not None:
                    self.loop.close()
            if "tree_structure" in obj:
                self.nodes.append(self.buffer)
                if self.loop is not None:
                    self.loop.update(len(self.nodes))
                    if len(self.nodes) % 10 == 0:
                        self.loop.set_description(
                            "dump_booster: %d/%d trees, %d nodes"
                            % (len(self.nodes), self.n_trees, self.stored)
                        )
                self.buffer = []
            if "decision_type" in obj:
                self.buffer.append(obj)
                self.stored += 1
            return obj

    if verbose >= 2:
        print("[dump_booster_model] to_json")
    info = {}
    ret = json.loads(
        string_buffer.value.decode("utf-8"),
        cls=Hook,
        info=info,
        n_trees=self.num_trees(),
        verbose=verbose,
    )
    ret["pandas_categorical"] = json.loads(
        json.dumps(self.pandas_categorical, default=jdwn)
    )
    if verbose >= 2:
        print("[dump_booster_model] end.")
    return ret, info


def _split_tree_ensemble_atts(attrs, split):
    """
    Splits the attributes of a TreeEnsembleRegressor into
    multiple trees in order to do the summation in double instead of floats.
    """
    trees_id = list(sorted(set(attrs["nodes_treeids"])))
    results = []
    index = 0
    while index < len(trees_id):
        index2 = min(index + split, len(trees_id))
        subset = set(trees_id[index:index2])

        indices_node = []
        indices_target = []
        for j, v in enumerate(attrs["nodes_treeids"]):
            if v in subset:
                indices_node.append(j)
        for j, v in enumerate(attrs["target_treeids"]):
            if v in subset:
                indices_target.append(j)

        if len(indices_node) >= len(attrs["nodes_treeids"]) or len(
            indices_target
        ) >= len(attrs["target_treeids"]):
            raise RuntimeError(  # pragma: no cover
                "Initial attributes are not consistant."
                "\nindex=%r index2=%r subset=%r"
                "\nnodes_treeids=%r\ntarget_treeids=%r"
                "\nindices_node=%r\nindices_target=%r"
                % (
                    index,
                    index2,
                    subset,
                    attrs["nodes_treeids"],
                    attrs["target_treeids"],
                    indices_node,
                    indices_target,
                )
            )

        ats = {}
        for name, att in attrs.items():
            if name == "nodes_treeids":
                new_att = [att[i] for i in indices_node]
                new_att = [i - att[0] for i in new_att]
            elif name == "target_treeids":
                new_att = [att[i] for i in indices_target]
                new_att = [i - att[0] for i in new_att]
            elif name.startswith("nodes_"):
                new_att = [att[i] for i in indices_node]
                assert len(new_att) == len(indices_node)
            elif name.startswith("target_"):
                new_att = [att[i] for i in indices_target]
                assert len(new_att) == len(indices_target)
            elif name == "name":
                new_att = "%s%d" % (att, len(results))
            else:
                new_att = att
            ats[name] = new_att

        results.append(ats)
        index = index2

    return results


def convert_lightgbm(scope, operator, container):
    """
    Converters for *lightgbm*.
    """
    verbose = getattr(container, "verbose", 0)
    gbm_model = operator.raw_operator
    gbm_text, info = dump_booster_model(gbm_model.booster_, verbose=verbose)
    modify_tree_for_rule_in_set(gbm_text, use_float=True, verbose=verbose, info=info)

    attrs = get_default_tree_classifier_attribute_pairs()
    attrs["name"] = operator.full_name

    # Create different attributes for classifier and
    # regressor, respectively
    post_transform = None
    if gbm_text["objective"].startswith("binary"):
        n_classes = 1
        attrs["post_transform"] = "LOGISTIC"
    elif gbm_text["objective"].startswith("multiclass"):
        n_classes = gbm_text["num_class"]
        attrs["post_transform"] = "SOFTMAX"
    elif gbm_text["objective"].startswith(("regression", "quantile")):
        n_classes = 1  # Regressor has only one output variable
        attrs["post_transform"] = "NONE"
        attrs["n_targets"] = n_classes
    elif gbm_text["objective"].startswith(("poisson", "gamma")):
        n_classes = 1  # Regressor has only one output variable
        attrs["n_targets"] = n_classes
        # 'Exp' is not a supported post_transform value in the ONNX spec yet,
        # so we need to add an 'Exp' post transform node to the model
        attrs["post_transform"] = "NONE"
        post_transform = "Exp"
    else:
        raise RuntimeError(
            "LightGBM objective should be cleaned already not '{}'.".format(
                gbm_text["objective"]
            )
        )

    # Use the same algorithm to parse the tree
    for i, tree in enumerate(gbm_text["tree_info"]):
        tree_id = i
        class_id = tree_id % n_classes
        # tree['shrinkage'] --> LightGbm provides figures with it already.
        learning_rate = 1.0
        _parse_tree_structure(
            tree_id, class_id, learning_rate, tree["tree_structure"], attrs
        )

    # Sort nodes_* attributes. For one tree, its node indexes
    # should appear in an ascent order in nodes_nodeids. Nodes
    # from a tree with a smaller tree index should appear
    # before trees with larger indexes in nodes_nodeids.
    node_numbers_per_tree = Counter(attrs["nodes_treeids"])
    tree_number = len(node_numbers_per_tree.keys())
    accumulated_node_numbers = [0] * tree_number
    for i in range(1, tree_number):
        accumulated_node_numbers[i] = (
            accumulated_node_numbers[i - 1] + node_numbers_per_tree[i - 1]
        )
    global_node_indexes = []
    for i in range(len(attrs["nodes_nodeids"])):
        tree_id = attrs["nodes_treeids"][i]
        node_id = attrs["nodes_nodeids"][i]
        global_node_indexes.append(accumulated_node_numbers[tree_id] + node_id)
    for k, v in attrs.items():
        if k.startswith("nodes_"):
            merged_indexes = zip(copy.deepcopy(global_node_indexes), v)
            sorted_list = [
                pair[1] for pair in sorted(merged_indexes, key=lambda x: x[0])
            ]
            attrs[k] = sorted_list

    # Create ONNX object
    if gbm_text["objective"].startswith("binary") or gbm_text["objective"].startswith(
        "multiclass"
    ):
        # Prepare label information for both of TreeEnsembleClassifier
        class_type = onnx_proto.TensorProto.STRING
        if all(
            isinstance(i, (numbers.Real, bool, np.bool_)) for i in gbm_model.classes_
        ):
            class_type = onnx_proto.TensorProto.INT64
            class_labels = [int(i) for i in gbm_model.classes_]
            attrs["classlabels_int64s"] = class_labels
        elif all(isinstance(i, str) for i in gbm_model.classes_):
            class_labels = [str(i) for i in gbm_model.classes_]
            attrs["classlabels_strings"] = class_labels
        else:
            raise ValueError("Only string and integer class labels are allowed")

        # Create tree classifier
        probability_tensor_name = scope.get_unique_variable_name("probability_tensor")
        label_tensor_name = scope.get_unique_variable_name("label_tensor")

        # onnx does not support int and float values for a float tensor
        update = {}
        for k, v in attrs.items():
            if not isinstance(v, list):
                continue
            tps = set(map(type, v))
            if len(tps) == 2:
                if tps == {int, float}:
                    update[k] = [float(x) for x in v]
        attrs.update(update)

        container.add_node(
            "TreeEnsembleClassifier",
            operator.input_full_names,
            [label_tensor_name, probability_tensor_name],
            op_domain="ai.onnx.ml",
            **attrs
        )

        prob_tensor = probability_tensor_name

        if gbm_model.boosting_type == "rf":
            col_index_name = scope.get_unique_variable_name("col_index")
            first_col_name = scope.get_unique_variable_name("first_col")
            zeroth_col_name = scope.get_unique_variable_name("zeroth_col")
            denominator_name = scope.get_unique_variable_name("denominator")
            modified_first_col_name = scope.get_unique_variable_name(
                "modified_first_col"
            )
            unit_float_tensor_name = scope.get_unique_variable_name("unit_float_tensor")
            merged_prob_name = scope.get_unique_variable_name("merged_prob")
            predicted_label_name = scope.get_unique_variable_name("predicted_label")
            classes_name = scope.get_unique_variable_name("classes")
            final_label_name = scope.get_unique_variable_name("final_label")

            container.add_initializer(
                col_index_name, onnx_proto.TensorProto.INT64, [], [1]
            )
            container.add_initializer(
                unit_float_tensor_name, onnx_proto.TensorProto.FLOAT, [], [1.0]
            )
            container.add_initializer(
                denominator_name, onnx_proto.TensorProto.FLOAT, [], [100.0]
            )
            container.add_initializer(
                classes_name, class_type, [len(class_labels)], class_labels
            )

            container.add_node(
                "ArrayFeatureExtractor",
                [probability_tensor_name, col_index_name],
                first_col_name,
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
                op_domain="ai.onnx.ml",
            )
            apply_div(
                scope,
                [first_col_name, denominator_name],
                modified_first_col_name,
                container,
                broadcast=1,
            )
            apply_sub(
                scope,
                [unit_float_tensor_name, modified_first_col_name],
                zeroth_col_name,
                container,
                broadcast=1,
            )
            container.add_node(
                "Concat",
                [zeroth_col_name, modified_first_col_name],
                merged_prob_name,
                name=scope.get_unique_operator_name("Concat"),
                axis=1,
            )
            container.add_node(
                "ArgMax",
                merged_prob_name,
                predicted_label_name,
                name=scope.get_unique_operator_name("ArgMax"),
                axis=1,
            )
            container.add_node(
                "ArrayFeatureExtractor",
                [classes_name, predicted_label_name],
                final_label_name,
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
                op_domain="ai.onnx.ml",
            )
            apply_reshape(
                scope,
                final_label_name,
                operator.outputs[0].full_name,
                container,
                desired_shape=[
                    -1,
                ],
            )
            prob_tensor = merged_prob_name
        else:
            container.add_node(
                "Identity",
                label_tensor_name,
                operator.outputs[0].full_name,
                name=scope.get_unique_operator_name("Identity"),
            )

        # Convert probability tensor to probability map
        # (keys are labels while values are the associated probabilities)
        container.add_node("Identity", prob_tensor, operator.outputs[1].full_name)
    else:
        # Create tree regressor
        output_name = scope.get_unique_variable_name("output")

        keys_to_be_renamed = list(k for k in attrs if k.startswith("class_"))

        for k in keys_to_be_renamed:
            # Rename class_* attribute to target_*
            # because TreeEnsebmleClassifier
            # and TreeEnsembleClassifier have different ONNX attributes
            attrs["target" + k[5:]] = copy.deepcopy(attrs[k])
            del attrs[k]

        # onnx does not support int and float values for a float tensor
        update = {}
        for k, v in attrs.items():
            if not isinstance(v, list):
                continue
            tps = set(map(type, v))
            if len(tps) == 2:
                if tps == {int, float}:
                    update[k] = [float(x) for x in v]
        attrs.update(update)

        split = getattr(operator, "split", None)
        if split in (None, -1):
            container.add_node(
                "TreeEnsembleRegressor",
                operator.input_full_names,
                output_name,
                op_domain="ai.onnx.ml",
                **attrs
            )
        else:
            tree_attrs = _split_tree_ensemble_atts(attrs, split)
            tree_nodes = []
            for i, ats in enumerate(tree_attrs):
                tree_name = scope.get_unique_variable_name("tree%d" % i)
                container.add_node(
                    "TreeEnsembleRegressor",
                    operator.input_full_names,
                    tree_name,
                    op_domain="ai.onnx.ml",
                    **ats
                )
                cast_name = scope.get_unique_variable_name("dtree%d" % i)
                container.add_node(
                    "Cast",
                    tree_name,
                    cast_name,
                    to=TensorProto.DOUBLE,  # pylint: disable=E1101
                    name=scope.get_unique_operator_name("dtree%d" % i),
                )
                tree_nodes.append(cast_name)
            cast_name = scope.get_unique_variable_name("ftrees")
            container.add_node(
                "Sum",
                tree_nodes,
                cast_name,
                name=scope.get_unique_operator_name("sumtree%d" % len(tree_nodes)),
            )
            container.add_node(
                "Cast",
                cast_name,
                output_name,
                to=TensorProto.FLOAT,  # pylint: disable=E1101
                name=scope.get_unique_operator_name("dtree%d" % i),
            )
        if gbm_model.boosting_type == "rf":
            denominator_name = scope.get_unique_variable_name("denominator")

            container.add_initializer(
                denominator_name, onnx_proto.TensorProto.FLOAT, [], [100.0]
            )

            apply_div(
                scope,
                [output_name, denominator_name],
                operator.output_full_names,
                container,
                broadcast=1,
            )
        elif post_transform:
            container.add_node(
                post_transform,
                output_name,
                operator.output_full_names,
                name=scope.get_unique_operator_name(post_transform),
            )
        else:
            container.add_node(
                "Identity",
                output_name,
                operator.output_full_names,
                name=scope.get_unique_operator_name("Identity"),
            )


def modify_tree_for_rule_in_set(
    gbm, use_float=False, verbose=0, count=0, info=None  # pylint: disable=R1710
):
    """
    LightGBM produces sometimes a tree with a node set
    to use rule ``==`` to a set of values (= in set),
    the values are separated by ``||``.
    This function unfold theses nodes.

    :param gbm: a tree coming from lightgbm dump
    :param use_float: use float otherwise int first
        then float if it does not work
    :param verbose: verbosity, use *tqdm* to show progress
    :param count: number of nodes already changed (origin) before this call
    :param info: addition information to speed up this search
    :return: number of changed nodes (include *count*)
    """
    if "tree_info" in gbm:
        if info is not None:
            dec_nodes = info["decision_nodes"]
        else:
            dec_nodes = None
        if verbose >= 2 and has_tqdm():
            from tqdm import tqdm

            loop = tqdm(gbm["tree_info"])
            for i, tree in enumerate(loop):
                loop.set_description("rules tree %d c=%d" % (i, count))
                count = modify_tree_for_rule_in_set(
                    tree,
                    use_float=use_float,
                    count=count,
                    info=None if dec_nodes is None else dec_nodes[i],
                )
        else:
            for i, tree in enumerate(gbm["tree_info"]):
                count = modify_tree_for_rule_in_set(
                    tree,
                    use_float=use_float,
                    count=count,
                    info=None if dec_nodes is None else dec_nodes[i],
                )
        return count

    if "tree_structure" in gbm:
        return modify_tree_for_rule_in_set(
            gbm["tree_structure"], use_float=use_float, count=count, info=info
        )

    if "decision_type" not in gbm:
        return count

    def str2number(val):
        if use_float:
            return float(val)
        else:
            try:
                return int(val)
            except ValueError:  # pragma: no cover
                return float(val)

    if info is None:

        def recursive_call(this, c):
            if "left_child" in this:
                c = process_node(this["left_child"], count=c)
            if "right_child" in this:
                c = process_node(this["right_child"], count=c)
            return c

        def process_node(node, count):
            if "decision_type" not in node:
                return count
            if node["decision_type"] != "==":
                return recursive_call(node, count)
            th = node["threshold"]
            if not isinstance(th, str):
                return recursive_call(node, count)
            pos = th.find("||")
            if pos == -1:
                return recursive_call(node, count)
            th1 = str2number(th[:pos])

            def doit():
                rest = th[pos + 2 :]
                if "||" not in rest:
                    rest = str2number(rest)

                node["threshold"] = th1
                new_node = node.copy()
                node["right_child"] = new_node
                new_node["threshold"] = rest

            doit()
            return recursive_call(node, count + 1)

        return process_node(gbm, count)

    # when info is used

    def split_node(node, th, pos):
        th1 = str2number(th[:pos])

        rest = th[pos + 2 :]
        if "||" not in rest:
            rest = str2number(rest)
            app = False
        else:
            app = True

        node["threshold"] = th1
        new_node = node.copy()
        node["right_child"] = new_node
        new_node["threshold"] = rest
        return new_node, app

    stack = deque(info)
    while len(stack) > 0:
        node = stack.pop()

        if "decision_type" not in node:
            continue  # leave

        if node["decision_type"] != "==":
            continue

        th = node["threshold"]
        if not isinstance(th, str):
            continue

        pos = th.find("||")
        if pos == -1:
            continue

        new_node, app = split_node(node, th, pos)
        count += 1
        if app:
            stack.append(new_node)

    return count


def convert_lgbm_zipmap(scope, operator, container):
    zipmap_attrs = {"name": scope.get_unique_operator_name("ZipMap")}
    if hasattr(operator, "classlabels_int64s"):
        zipmap_attrs["classlabels_int64s"] = operator.classlabels_int64s
        to_type = onnx_proto.TensorProto.INT64
    elif hasattr(operator, "classlabels_strings"):
        zipmap_attrs["classlabels_strings"] = operator.classlabels_strings
        to_type = onnx_proto.TensorProto.STRING
    else:
        raise RuntimeError("Unknown class type.")
    if to_type == onnx_proto.TensorProto.STRING:
        apply_identity(
            scope,
            operator.inputs[0].full_name,
            operator.outputs[0].full_name,
            container,
        )
    else:
        apply_cast(
            scope,
            operator.inputs[0].full_name,
            operator.outputs[0].full_name,
            container,
            to=to_type,
        )

    if operator.zipmap:
        container.add_node(
            "ZipMap",
            operator.inputs[1].full_name,
            operator.outputs[1].full_name,
            op_domain="ai.onnx.ml",
            **zipmap_attrs
        )
    else:
        # onnxconverter-common when trying to remove identity nodes
        # if node identity is used.
        one = scope.get_unique_variable_name("one")

        container.add_initializer(one, onnx_proto.TensorProto.FLOAT, [], [1])
        container.add_node(
            "Mul", [operator.inputs[1].full_name, one], operator.outputs[1].full_name
        )


register_converter("LgbmClassifier", convert_lightgbm)
register_converter("LgbmRegressor", convert_lightgbm)
register_converter("LgbmZipMap", convert_lgbm_zipmap)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_identity(scope, operator, container):
    container.add_node(
        "Identity",
        operator.input_full_names,
        operator.output_full_names,
        name=operator.full_name,
    )


register_converter("identity", convert_identity)



# SPDX-License-Identifier: Apache-2.0

from ...common.data_types import StringTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_tokenizer(scope, operator, container):
    op = operator.raw_operator
    if op is None:
        raise RuntimeError("op cannot be None.")
    # the SPARK version converts text to lowercase and applies "\\s" regexp to it
    # Here we'll tokenize and then normalize (to convert to lowercase)
    lowercase_output = scope.get_unique_variable_name("lowercase_tensor")
    container.add_node(
        "StringNormalizer",
        operator.input_full_names[0],
        lowercase_output,
        op_domain="ai.onnx",
        name=scope.get_unique_operator_name("StringNormalizer"),
        op_version=10,
        case_change_action="LOWER",
    )
    container.add_node(
        "Tokenizer",
        lowercase_output,
        operator.output_full_names,
        op_domain="com.microsoft",
        name=scope.get_unique_operator_name("Tokenizer"),
        op_version=1,
        mark=0,
        separators=[" ", "\t", "\r", "\n"],
        pad_value="##ERROR##",
        mincharnum=1,
    )


register_converter("pyspark.ml.feature.Tokenizer", convert_sparkml_tokenizer)


def calculate_sparkml_tokenizer_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[StringTensorType])
    operator.outputs[0].type = StringTensorType()


register_shape_calculator(
    "pyspark.ml.feature.Tokenizer", calculate_sparkml_tokenizer_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from .....proto import onnx_proto
from ....common._registration import register_converter
from .Reshape import apply_reshape
from .SimpleRNN import extract_rnn_activation_info


def convert_gru(scope, operator, container):
    # The GRU inputs are feature vector, X, and initial hidden state, h_init. Let C_in and C_out denote
    # the input feature length and the output dimension, respectively. Assume that S is the sequence length
    # (i.e., S-axis is the time axis of a sequence). In CorML, the shapes of X and h_init are [S, C_in] and
    # [1, C_out] respectively. In ONNX, the two shapes become [S, N, C_in] (X) and [D, N, C_out]
    # (h_init), where N is the batch size and S is the sequence length (i.e., S-axis is the time axis). To
    # simulate CoreML GRU under ONNX, we need to introduce some extra operators. Note that N=1 and D=1 always
    # hold in ONNX if we are considering GRU from CoreML because there is no batch size in CoreML and CoreML GRU
    # is always uni-directional.
    #
    # Below we provide a visualization of our conversion for CoreML GRU.
    #
    # Symbols:
    #
    #  X: input feature vector of CoreML GRU
    #  h_init: initial GRU state of CoreML
    #  Y: CoreML GRU's output. It can be [S, C_out] (if sequence_output is on ) or [1, C_out] (if sequence_output is off)
    #  Y_h: CoreML GRU's last hidden state
    #
    #  X': input features of ONNX GRU
    #  h_init': initial GRU state of ONNX
    #  Y': ONNX GRU's output
    #  Y_h': ONNX GRU's last hidden state
    #
    # Computational graph of CoreML GRU (sequence_output is on):
    #
    # X [S, C_in] ---> CoreML GRU ---> Y [S, C_out]
    #                    ^     |
    #                    |     |
    # h_init [1, C_cou] -'     '---> Y_h [1, C_out]
    #
    # Computational graph we use to repreent CoreML GRU into ONNX (sequence_output is on):
    #
    # X [S, C_in] --> Reshape --> X' [S, 1, C_in] --> ONNX GRU --> Y' [S, 1, C_out] --> Reshape --> Y [S, C_out]
    #                                                    ^ |
    #                                                    | |
    # h_init [1, C_out] --> Reshape -> h_init' [1, 1, C]-' '--> Y_h' [1, 1, C_out] ---> Reshape --> Y_h [1, C_out]
    #
    # Computational graph of CoreML GRU (sequence_output is off):
    #
    # X [S, C_in] ---> CoreML GRU ---> Y [1, C_out]
    #                    ^     |
    #                    |     |
    # h_init [1, C_cou] -'     '---> Y_h [1, C_out] Note that in this case, Y=Y_h.
    #
    # Computational graph we use to represent CoreML GRU into ONNX (sequence_output is off):
    #
    # X [S, C_in] --> Reshape --> X' [S, 1, C_in] --> ONNX GRU --> Y' [S, 1, C_out] (this output won't be connected with others)
    #                                                    ^ |
    #                                                    | |
    # h_init [1, C_out] --> Reshape -> h_init' [1, 1, C]-' '--> Y_h' [1, 1, C_out] ---> Reshape --> Y [1, C_out]
    #                                                                                                  |
    #                                                                                                  v
    #                                                                                               Identity
    #                                                                                                  |
    #                                                                                                  v
    #                                                                                             Y_h [1, C_out]

    params = operator.raw_operator.gru
    input_size = params.inputVectorSize
    hidden_size = params.outputVectorSize

    # Initialize GRU's attributes. They will be used to build GRU in the end of this function.
    gru_op_name = scope.get_unique_operator_name("GRU")
    gru_attrs = {"name": gru_op_name}
    gru_inputs = []
    gru_outputs = []

    # Resahpe CoreML variable into ONNX format for feeding it into ONNX GRU
    gru_x_reshape_name = scope.get_unique_variable_name(gru_op_name + "_X_reshape")
    apply_reshape(
        scope,
        operator.inputs[0].full_name,
        gru_x_reshape_name,
        container,
        desired_shape=[-1, 1, input_size],
    )
    gru_inputs.append(gru_x_reshape_name)

    # Create weight matrices of GRU and add it into ONNX GRU's input list
    matrices_w = np.concatenate(
        [
            params.updateGateWeightMatrix.floatValue,
            params.resetGateWeightMatrix.floatValue,
            params.outputGateWeightMatrix.floatValue,
        ]
    )
    matrices_w_name = scope.get_unique_variable_name(gru_op_name + "_W")
    container.add_initializer(
        matrices_w_name,
        onnx_proto.TensorProto.FLOAT,
        [1, 3 * hidden_size, input_size],
        matrices_w,
    )
    gru_inputs.append(matrices_w_name)

    # Create recursion matrices of GRU and add it into ONNX GRU's input list
    matrices_r = np.concatenate(
        [
            params.updateGateRecursionMatrix.floatValue,
            params.resetGateRecursionMatrix.floatValue,
            params.outputGateRecursionMatrix.floatValue,
        ]
    )
    matrices_r_name = scope.get_unique_variable_name(gru_op_name + "_R")
    container.add_initializer(
        matrices_r_name,
        onnx_proto.TensorProto.FLOAT,
        [1, 3 * hidden_size, hidden_size],
        matrices_r,
    )
    gru_inputs.append(matrices_r_name)

    if params.hasBiasVectors:
        # Create bias vectors of GRU and add them into ONNX GRU's input list
        vectors_b = np.concatenate(
            [
                params.updateGateBiasVector.floatValue,
                params.resetGateBiasVector.floatValue,
                params.outputGateBiasVector.floatValue,
                np.zeros(3 * hidden_size),
            ]
        )
        vectors_b_name = scope.get_unique_variable_name(gru_op_name + "_B")
        container.add_initializer(
            vectors_b_name,
            onnx_proto.TensorProto.FLOAT,
            [1, 6 * hidden_size],
            vectors_b,
        )
        gru_inputs.append(vectors_b_name)
    else:
        # Because operator's arguments are position-sensitive, we need an empty string even if
        # this variable doesn't exist.
        gru_inputs.append("")

    # The argument, sequence length, is always missing when converting CoreML GRU.
    gru_inputs.append("")

    # Handle initial hidden state if it exists
    if len(operator.inputs) == 2:
        # Change the shape of initial state in CoreML so that ONNX's GRU is willing to take it.
        gru_h_init_reshape_name = scope.get_unique_variable_name(
            gru_op_name + "_h_init"
        )
        apply_reshape(
            scope,
            operator.inputs[1].full_name,
            gru_h_init_reshape_name,
            container,
            desired_shape=[1, 1, hidden_size],
        )
        gru_inputs.append(gru_h_init_reshape_name)

        # Add a zero initializer to initial hidden state so that this variable becomes optional
        container.add_initializer(
            operator.inputs[1].full_name,
            onnx_proto.TensorProto.FLOAT,
            operator.inputs[1].type.shape,
            np.zeros(shape=operator.inputs[1].type.shape).flatten(),
        )
    else:
        # Because operator's arguments are position-sensitive, we need an empty string even if
        # this variable doesn't exist.
        gru_inputs.append("")

    activation_types = []
    alphas = []
    betas = []
    for activation in params.activations:
        activation_type, alpha, beta = extract_rnn_activation_info(activation)
        activation_types.append(activation_type.encode("utf-8"))
        if alpha is not None:
            alphas.append(alpha)
        if beta is not None:
            betas.append(beta)
    gru_attrs["activations"] = activation_types
    if alphas:
        gru_attrs["activation_alpha"] = alphas
    if betas:
        gru_attrs["activation_beta"] = betas
    gru_attrs["direction"] = "reverse" if params.reverseInput else "forward"
    gru_attrs["hidden_size"] = hidden_size

    # Set up version-dependent attributes
    if container.target_opset < 5:
        gru_attrs["output_sequence"] = params.sequenceOutput
        op_version = 1
    elif container.target_opset < 7:
        gru_attrs["linear_before_reset"] = 0
        gru_attrs["output_sequence"] = params.sequenceOutput
        op_version = 3
    else:
        gru_attrs["linear_before_reset"] = 0
        op_version = 7

    # Create the major GRU operator in ONNX.
    gru_y_name = scope.get_unique_variable_name(gru_op_name + "_Y")
    gru_y_h_name = scope.get_unique_variable_name(gru_op_name + "_Y_h")
    gru_outputs.extend([gru_y_name, gru_y_h_name])
    container.add_node(
        "GRU", gru_inputs, gru_outputs, op_version=op_version, **gru_attrs
    )

    # To simulate CoreML LSTM, we add post-processing
    # operators to adjust ONNX LSTM outputs
    if params.sequenceOutput:
        # Again, the output shapes in ONNX's GRU
        # is not consistent with that in CoreML, so we need
        # to adjust the result produced by ONNX according to CoreML format.
        apply_reshape(
            scope,
            gru_y_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[-1, hidden_size],
        )

        # Handle the second output, the last hidden state of a sequence, if exists.
        if len(operator.outputs) == 2:
            apply_reshape(
                scope,
                gru_y_h_name,
                operator.outputs[1].full_name,
                container,
                desired_shape=[1, hidden_size],
            )
    else:
        # Recall that when sequence output is false, the first and the second outputs of GRU
        # are identical. Thus, we can ignore ONNX GRU's first output.
        apply_reshape(
            scope,
            gru_y_h_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[1, hidden_size],
        )

        if len(operator.outputs) == 2:
            container.add_node(
                "Identity",
                operator.outputs[0].full_name,
                operator.outputs[1].full_name,
                name=scope.get_unique_operator_name("Identity"),
            )


register_converter("gru", convert_gru)



# SPDX-License-Identifier: Apache-2.0

"""
Utility functions for Spark ML to Onnx conversion intended for the end user mainly
"""
from ..common.data_types import StringTensorType, FloatTensorType


def buildInitialTypesSimple(dataframe):
    types = []
    for field in dataframe.schema.fields:
        types.append((field.name, getTensorTypeFromSpark(str(field.dataType))))
    return types


def getTensorTypeFromSpark(sparktype):
    if sparktype in ("StringType", "StringType()"):
        return StringTensorType([1, 1])
    if (
        sparktype == "DecimalType"
        or sparktype == "DecimalType()"
        or sparktype == "DoubleType"
        or sparktype == "DoubleType()"
        or sparktype == "FloatType"
        or sparktype == "FloatType()"
        or sparktype == "LongType"
        or sparktype == "LongType()"
        or sparktype == "IntegerType"
        or sparktype == "IntegerType()"
        or sparktype == "ShortType"
        or sparktype == "ShortType()"
        or sparktype == "ByteType"
        or sparktype == "ByteType()"
        or sparktype == "BooleanType"
        or sparktype == "BooleanType()"
    ):
        return FloatTensorType([None, 1])
    raise TypeError(f"Cannot map this type to Onnx types: {sparktype}.")


def buildInputDictSimple(dataframe):
    import numpy

    result = {}
    for field in dataframe.schema.fields:
        if str(field.dataType) in ("StringType", "StringType()"):
            result[field.name] = dataframe.select(field.name).toPandas().values
        else:
            result[field.name] = (
                dataframe.select(field.name).toPandas().values.astype(numpy.float32)
            )
    return result


class SparkMlConversionError(Exception):
    pass



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.shape_calculator import calculate_linear_regressor_output_shapes

register_shape_calculator("XGBRegressor", calculate_linear_regressor_output_shapes)
register_shape_calculator("XGBRFRegressor", calculate_linear_regressor_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import json
import re
import pprint
from packaging.version import Version
import numpy as np
from xgboost import XGBRegressor, XGBClassifier, __version__

try:
    from xgboost import XGBRFRegressor, XGBRFClassifier
except ImportError:
    # old version of xgboost
    XGBRFRegressor, XGBRFClassifier = None, None
from onnxconverter_common.data_types import FloatTensorType
from ..common._container import XGBoostModelContainer
from ..common._topology import Topology


xgboost_classifier_list = [XGBClassifier]

# Associate types with our operator names.
xgboost_operator_name_map = {
    XGBClassifier: "XGBClassifier",
    XGBRegressor: "XGBRegressor",
}

if XGBRFClassifier:
    xgboost_operator_name_map.update(
        {
            XGBRFClassifier: "XGBRFClassifier",
            XGBRFRegressor: "XGBRFRegressor",
        }
    )
    xgboost_classifier_list.append(XGBRFClassifier)


def _append_covers(node):
    res = []
    if "cover" in node:
        res.append(node["cover"])
    if "children" in node:
        for ch in node["children"]:
            res.extend(_append_covers(ch))
    return res


def _get_attributes(booster):
    atts = booster.attributes()
    dp = booster.get_dump(dump_format="json", with_stats=True)
    res = [json.loads(d) for d in dp]

    # num_class
    if Version(__version__) < Version("1.5"):
        state = booster.__getstate__()
        bstate = bytes(state["handle"])
        reg = re.compile(b'("tree_info":\\[[0-9,]*\\])')
        objs = list(set(reg.findall(bstate)))
        if len(objs) != 1:
            raise RuntimeError(
                "Unable to retrieve the tree coefficients from\n%s"
                "" % bstate.decode("ascii", errors="ignore")
            )
        tree_info = json.loads("{{{}}}".format(objs[0].decode("ascii")))["tree_info"]
        num_class = len(set(tree_info))
        trees = len(res)
        try:
            ntrees = booster.best_ntree_limit
        except AttributeError:
            ntrees = trees // num_class if num_class > 0 else trees
    else:
        trees = len(res)
        ntrees = getattr(booster, "best_ntree_limit", trees)
        config = json.loads(booster.save_config())["learner"]["learner_model_param"]
        num_class = int(config["num_class"]) if "num_class" in config else 0
        if num_class == 0 and ntrees > 0:
            num_class = trees // ntrees
        if num_class == 0:
            raise RuntimeError(
                f"Unable to retrieve the number of classes, num_class={num_class}, "
                f"trees={trees}, ntrees={ntrees}, config={config}."
            )

    kwargs = atts.copy()
    kwargs["feature_names"] = booster.feature_names
    kwargs["n_estimators"] = ntrees

    # covers
    covs = []
    for tr in res:
        covs.extend(_append_covers(tr))

    if all(map(lambda x: int(x) == x, set(covs))):
        # regression
        kwargs["num_target"] = num_class
        kwargs["num_class"] = 0
        kwargs["objective"] = "reg:squarederror"
    else:
        # classification
        kwargs["num_class"] = num_class
        if num_class != 1:
            if Version(__version__) < Version("1.5"):
                reg = re.compile(b"(multi:[a-z]{1,15})")
                objs = list(set(reg.findall(bstate)))
                if len(objs) == 1:
                    kwargs["objective"] = objs[0].decode("ascii")
                else:
                    raise RuntimeError(
                        "Unable to guess objective in %r (trees=%r, ntrees=%r, num_class=%r)"
                        "." % (objs, trees, ntrees, kwargs["num_class"])
                    )
            else:
                att = json.loads(booster.save_config())
                kwargs["objective"] = att["learner"]["objective"]["name"]
                nc = int(att["learner"]["learner_model_param"]["num_class"])
                if nc != num_class:
                    raise RuntimeError(
                        "Mismatched value %r != %r from\n%s"
                        % (nc, num_class, pprint.pformat(att))
                    )
        else:
            kwargs["objective"] = "binary:logistic"

    if "base_score" not in kwargs:
        kwargs["base_score"] = 0.5
    elif isinstance(kwargs["base_score"], str):
        kwargs["base_score"] = float(kwargs["base_score"])
    return kwargs


class WrappedBooster:
    def __init__(self, booster):
        self.booster_ = booster
        self.kwargs = _get_attributes(booster)

        if self.kwargs["num_class"] > 0:
            self.classes_ = self._generate_classes(self.kwargs)
            self.operator_name = "XGBClassifier"
        else:
            self.operator_name = "XGBRegressor"

    def get_xgb_params(self):
        return {k: v for k, v in self.kwargs.items() if v is not None}

    def get_booster(self):
        return self.booster_

    def _generate_classes(self, model_dict):
        if model_dict["num_class"] == 1:
            return np.asarray([0, 1])
        return np.arange(model_dict["num_class"])


def _get_xgboost_operator_name(model):
    """
    Get operator name of the input argument

    :param model_type:  A xgboost object.
    :return: A string which stands for the type of the input model in our conversion framework
    """
    if isinstance(model, WrappedBooster):
        return model.operator_name
    if type(model) not in xgboost_operator_name_map:
        raise ValueError("No proper operator name found for '%s'" % type(model))
    return xgboost_operator_name_map[type(model)]


def _parse_xgboost_simple_model(scope, model, inputs):
    """
    This function handles all non-pipeline models.

    :param scope: Scope object
    :param model: A xgboost object
    :param inputs: A list of variables
    :return: A list of output variables which will be passed to next stage
    """
    this_operator = scope.declare_local_operator(
        _get_xgboost_operator_name(model), model
    )
    this_operator.inputs = inputs

    if type(model) in xgboost_classifier_list or getattr(
        model, "operator_name", None
    ) in ("XGBClassifier", "XGBRFClassifier"):
        # For classifiers, we may have two outputs, one for label and
        # the other one for probabilities of all classes.
        # Notice that their types here are not necessarily correct
        # and they will be fixed in shape inference phase
        label_variable = scope.declare_local_variable("label", FloatTensorType())
        probability_map_variable = scope.declare_local_variable(
            "probabilities", FloatTensorType()
        )
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(probability_map_variable)
    else:
        # We assume that all scikit-learn operator can only produce a single float tensor.
        variable = scope.declare_local_variable("variable", FloatTensorType())
        this_operator.outputs.append(variable)
    return this_operator.outputs


def _parse_xgboost(scope, model, inputs):
    """
    This is a delegate function. It doesn't nothing but invoke
    the correct parsing function according to the input
    model's type.
    :param scope: Scope object
    :param model: A xgboost object
    :param inputs: A list of variables
    :return: The output variables produced by the input model
    """
    return _parse_xgboost_simple_model(scope, model, inputs)


def parse_xgboost(
    model,
    initial_types=None,
    target_opset=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    raw_model_container = XGBoostModelContainer(model)
    topology = Topology(
        raw_model_container,
        default_batch_size="None",
        initial_types=initial_types,
        target_opset=target_opset,
        custom_conversion_functions=custom_conversion_functions,
        custom_shape_calculators=custom_shape_calculators,
    )
    scope = topology.declare_scope("__root__")

    inputs = []
    for var_name, initial_type in initial_types:
        inputs.append(scope.declare_local_variable(var_name, initial_type))

    for variable in inputs:
        raw_model_container.add_input(variable)

    outputs = _parse_xgboost(scope, model, inputs)

    for variable in outputs:
        raw_model_container.add_output(variable)

    return topology



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_reduce(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import ReduceLayerParams as Params

    reduce_mode_table = {
        Params.SUM: "ReduceSum",
        Params.AVG: "ReduceMean",
        Params.PROD: "ReduceProd",
        Params.LOGSUM: "ReduceLogSum",
        Params.SUMSQUARE: "ReduceSumSquare",
        Params.L1: "ReduceL1",
        Params.L2: "ReduceL2",
        Params.MAX: "ReduceMax",
        Params.MIN: "ReduceMin",
        Params.ARGMAX: "ArgMax",
    }
    params = operator.raw_operator.reduce
    reduce_mode = reduce_mode_table[params.mode]
    reduce_name = scope.get_unique_operator_name(reduce_mode)
    # CoreML's reduce operator is used to process tensors
    # with shape [C, H, W]. Notice that [C, H, W] in CoreML
    # corresponds to [N, C, H, W] in ONNX because ONNX explicitly
    # get the batch axis. If a CoreML reduce is working
    # on CoreML's C-axis, the corresponding ONNX axis's
    # index would be 1 (for the 2nd axis in [N, C, H, W]-system).
    reduce_axis_table = {
        Params.CHW: [1, 2, 3],
        Params.HW: [2, 3],
        Params.C: [1],
        Params.H: [2],
        Params.W: [3],
    }
    reduce_axis = reduce_axis_table[params.axis]
    attrs = {"name": reduce_name, "axes": reduce_axis}
    container.add_node(
        reduce_mode, operator.input_full_names, operator.output_full_names, **attrs
    )


register_converter("reduce", convert_reduce)



# SPDX-License-Identifier: Apache-2.0

"""
Testcase Base class for SparkML tests
"""
import os
import inspect
import unittest
from tests.sparkml.sparkml_test_utils import start_spark, stop_spark


class SparkMlTestCase(unittest.TestCase):
    def _get_spark_options(self):
        return None

    def setUp(self):
        if os.name == "nt" and os.environ.get("HADOOP_HOME") is None:
            this_script_dir = os.path.dirname(
                os.path.abspath(inspect.getfile(inspect.currentframe()))
            )
            print("setting HADOOP_HOME to: ", this_script_dir)
            os.environ["HADOOP_HOME"] = this_script_dir
        self.spark = start_spark(self._get_spark_options())

    def tearDown(self):
        stop_spark(self.spark)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from .....proto import onnx_proto
from ....common._registration import register_converter
from .Reshape import apply_reshape


def extract_rnn_activation_info(activation):
    activation_type = activation.WhichOneof("NonlinearityType")
    alpha = None
    beta = None

    activation_map = {
        "linear": "Affine",
        "ReLU": "Relu",
        "leakyReLU": "LeakyRelu",
        "thresholdedReLU": "ThresholdedRelu",
        "PReLU": "PRelu",
        "tanh": "Tanh",
        "scaledTanh": "ScaledTanh",
        "sigmoid": "Sigmoid",
        "sigmoidHard": "HardSigmoid",
        "ELU": "Elu",
        "softsign": "Softsign",
        "softplus": "Softplus",
        "parametricSoftplus": "ParametricSoftplus",
    }

    if activation_type not in activation_map:
        raise ValueError("Unsupported activation function: {}".format(activation_type))

    # Notice that if we see a default vaulue (i.e., 0 for float), we may replace it with
    # the real default parameter for the specified activation function if necessary.
    if activation_type == "leakyReLU":
        alpha = activation.leakyReLU.alpha
        if alpha == 0:
            alpha = 0.3
    elif activation_type == "PReLU":
        raise RuntimeError(
            "Unsupported activation function: {}".format(activation_type)
        )
    elif activation_type == "ELU":
        alpha = activation.ELU.alpha
    elif activation_type == "thresholdedReLU":
        alpha = activation.thresholdedReLU.alpha
        if alpha == 0:
            alpha = 1.0
    elif activation_type == "scaledTanh":
        alpha = activation.scaledTanh.alpha
        beta = activation.scaledTanh.beta
    elif activation_type == "linear":
        alpha = activation.linear.alpha
        beta = activation.linear.beta
        if alpha == 0:
            alpha = 1.0
    elif activation_type == "sigmoidHard":
        alpha = activation.sigmoidHard.alpha
        beta = activation.sigmoidHard.beta
        if alpha == 0:
            alpha = 0.2
        if beta == 0:
            beta = 0.5
    elif activation_type == "parametricSoftplus":
        raise RuntimeError(
            "Unsupported activation function: {}".format(activation_type)
        )

    return activation_map[activation_type], alpha, beta


def convert_simple_rnn(scope, operator, container):
    # The RNN inputs are feature vector, X, and initial hidden state, h_init. Let C_in and C_out denote
    # the input feature length and the output dimension, respectively. Assume that S is the sequence length
    # (i.e., S-axis is the time axis of a sequence). In CorML, the shapes of X and h_init are [S, C_in] and
    # [1, C_out] respectively. In ONNX, the two shapes become [S, N, C_in] (X) and [D, N, C_out]
    # (h_init), where N is the batch size. To simulate CoreML RNN under ONNX, we need to introduce some extra
    # operators. Note that N=1 and D=1 always hold in ONNX if we are considering RNN from CoreML because there
    # is no batch size in CoreML and CoreML RNN is always uni-directional.
    #
    # Below we provide a visualization of our conversion for CoreML RNN.
    #
    # Symbols:
    #
    #  X: input features of CoreML RNN
    #  h_init: initial RNN state of CoreML
    #  Y: CoreML RNN's output. It can be [S, C_out] (if sequence_output is on ) or [1, C_out] (if sequence_output is off)
    #  Y_h: CoreML RNN's last hidden state
    #
    #  X': input features of ONNX RNN
    #  h_init': initial RNN state of ONNX
    #  Y': ONNX RNN's output
    #  Y_h': ONNX RNN's last hidden state
    #
    # Computational graph of CoreML RNN (sequence_output is on):
    #
    # X [S, C_in] ---> CoreML RNN ---> Y [S, C_out]
    #                    ^     |
    #                    |     |
    # h_init [1, C_out] -'     '---> Y_h [1, C_out]
    #
    # Computational graph we use for represent CoreML RNN into ONNX (sequence_output is on):
    #
    # X [S, C_in] --> Reshape --> X' [S, 1, C_in] -----> ONNX RNN --> Y' [S, 1, C_out] --> Reshape --> Y [S, C_out]
    #                                                        ^ |
    #                                                        | |
    # h_init [1, C_out] --> Reshape -> h_init' [1, 1, C_out]-' '--> Y_h' [1, 1, C_out] ---> Reshape --> Y_h [1, C_out]
    #
    # Computational graph of CoreML RNN (sequence_output is off):
    #
    # X [S, C_in] ---> CoreML RNN ---> Y [1, C_out]
    #                    ^     |
    #                    |     |
    # h_init [1, C_cou] -'     '---> Y_h [1, C_out] Note that in this case, Y=Y_h.
    #
    # Computational graph we use to represent CoreML RNN into ONNX (sequence_output is off):
    #
    # X [S, C_in] --> Reshape --> X' [S, 1, C_in] -----> ONNX RNN --> Y' [S, 1, C_out] Here Y' is useless.
    #                                                        ^ |
    #                                                        | |
    # h_init [1, C_out] --> Reshape -> h_init' [1, 1, C_out]-' '--> Y_h' [1, 1, C_out] ---> Reshape --> Y [1, C_out]
    #                                                                                                   |
    #                                                                                                   v
    #                                                                                                 Identity
    #                                                                                                   |
    #                                                                                                   v
    #                                                                                               Y_h [1, C_out]

    params = operator.raw_operator.simpleRecurrent
    input_size = params.inputVectorSize
    hidden_size = params.outputVectorSize

    X_name = operator.inputs[0].full_name
    X_reshape_name = scope.get_unique_variable_name("X")
    apply_reshape(
        scope, X_name, X_reshape_name, container, desired_shape=[-1, 1, input_size]
    )

    rnn_op_name = scope.get_unique_operator_name("RNN")
    rnn_attrs = {"name": rnn_op_name}
    rnn_inputs = [X_reshape_name]

    # Load RNN's weight matrix and add it into RNN's input list
    rnn_w_name = scope.get_unique_variable_name(rnn_op_name + "_W")
    container.add_initializer(
        rnn_w_name,
        onnx_proto.TensorProto.FLOAT,
        [1, hidden_size, input_size],
        params.weightMatrix.floatValue,
    )
    rnn_inputs.append(rnn_w_name)

    # Load RNN's recursion matrix and add it into RNN's input list
    rnn_r_name = scope.get_unique_variable_name(rnn_op_name + "_R")
    container.add_initializer(
        rnn_r_name,
        onnx_proto.TensorProto.FLOAT,
        [1, hidden_size, hidden_size],
        params.recursionMatrix.floatValue,
    )
    rnn_inputs.append(rnn_r_name)

    if params.hasBiasVector:
        # Load RNN's bias vector and add it into RNN's input list
        rnn_b_name = scope.get_unique_variable_name(rnn_op_name + "_B")
        rnn_b_content = np.concatenate(
            [params.biasVector.floatValue, np.zeros(hidden_size)]
        ).flatten()
        container.add_initializer(
            rnn_b_name,
            onnx_proto.TensorProto.FLOAT,
            [1, 2 * hidden_size],
            rnn_b_content,
        )
        rnn_inputs.append(rnn_b_name)
    else:
        # Input names are position-sensitive, so for optional
        # but missing inputs, we need to provide an empty string.
        rnn_inputs.append("")

    # The input, sequence_lens, in ONNX is alwasy optional for
    # this conversion, so here is always an empty string.
    rnn_inputs.append("")

    # If initial hidden state is provided, we add it into RNN's
    # input list after adjusting its shape.
    if len(operator.inputs) == 2:
        rnn_h_init_reshape_name = scope.get_unique_variable_name(
            rnn_op_name + "_h_init"
        )
        apply_reshape(
            scope,
            operator.inputs[1].full_name,
            rnn_h_init_reshape_name,
            container,
            desired_shape=[1, 1, hidden_size],
        )

        rnn_inputs.append(rnn_h_init_reshape_name)
        # Add a zero initializer to initial hidden state so that
        # this variable becomes optional
        container.add_initializer(
            operator.inputs[1].full_name,
            onnx_proto.TensorProto.FLOAT,
            operator.inputs[1].type.shape,
            np.zeros(shape=operator.inputs[1].type.shape).flatten(),
        )
    else:
        # Input names are position-sensitive, so for optional
        # but missing inputs, we need to provide an empty string.
        rnn_inputs.append("")

    # Add RNN's information of activation function
    activation, alpha, beta = extract_rnn_activation_info(params.activation)
    rnn_attrs["activations"] = [activation.encode("utf-8")]
    if alpha is not None:
        rnn_attrs["activation_alpha"] = [alpha]
    if beta is not None:
        rnn_attrs["activation_beta"] = [beta]

    # Set up other attributes
    rnn_attrs["direction"] = "reverse" if params.reverseInput else "forward"
    rnn_attrs["hidden_size"] = hidden_size

    # Set up version-dependent attributes
    if container.target_opset < 7:
        rnn_attrs["output_sequence"] = params.sequenceOutput
        op_version = 1
    else:
        op_version = 7

    # We use the collected information to build ONNX's RNN.
    # ONNX RNN's outputs will be saved onto two intermediate
    # tensors and we will adjust them subsequently to mimic
    # Keras output format.
    rnn_y_name = scope.get_unique_variable_name(rnn_op_name + "_Y")
    rnn_h_name = scope.get_unique_variable_name(rnn_op_name + "_Y_h")
    container.add_node(
        "RNN", rnn_inputs, [rnn_y_name, rnn_h_name], op_version=op_version, **rnn_attrs
    )

    # Set up outputs' of RNN
    if params.sequenceOutput:
        # Connect ONNX's output and CoreML's output via a reshape operator
        apply_reshape(
            scope,
            rnn_y_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[-1, hidden_size],
        )

        # Handel the second RNN output (aka last hidden state), which is optional.
        if len(operator.outputs) == 2:
            # Connect ONNX's output and CoreML's output via a reshape operator
            apply_reshape(
                scope,
                rnn_h_name,
                operator.outputs[1].full_name,
                container,
                desired_shape=[1, hidden_size],
            )
    else:
        # According to CoreML, its two outputs are always identical, so
        # we just need to compute one of them and produce
        # the other one using an identity operator. Note that the
        # first ONNX RNN output is undefined in this case.

        # Reshape last hidden state's ONNX format to its CoreML format
        apply_reshape(
            scope,
            rnn_h_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=[1, hidden_size],
        )

        if len(operator.outputs) == 2:
            # Copy the first output to the second output
            container.add_node(
                "Identity",
                operator.outputs[0].full_name,
                operator.outputs[1].full_name,
                name=scope.get_unique_operator_name("Identity"),
            )


register_converter("simpleRecurrent", convert_simple_rnn)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_upsample
from ....common._registration import register_converter


def convert_upsample(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import UpsampleLayerParams as Params

    params = operator.raw_operator.upsample

    if params.mode == Params.NN:
        mode = "NEAREST"
    elif params.mode == Params.BILINEAR:
        mode = "BILINEAR"
    else:
        raise ValueError("Unsupported interpolation model in up-sampling")

    width_scale = float(params.scalingFactor[1])
    height_scale = float(params.scalingFactor[0])

    apply_upsample(
        scope,
        operator.input_full_names[0],
        operator.output_full_names,
        container,
        operator_name=None,
        mode=mode,
        scales=[1, 1, height_scale, width_scale],
    )


register_converter("upsample", convert_upsample)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._registration import register_converter
from ....common._apply_operation import apply_add, apply_mul


def convert_preprocessing_scaler(scope, operator, container):
    params = operator.raw_operator.scaler
    # Specify some of this operator's attribute.
    # The scale parameter in CoreML is always a scalar.
    # We just copy it and let ONNX scaler to broadcast it to all channels.

    color_space = operator.inputs[0].type.color_space
    if color_space == "Gray8":
        bias = [params.grayBias]
    elif color_space == "Rgb8":
        bias = [params.redBias, params.greenBias, params.blueBias]
    elif color_space == "Bgr8":
        bias = [params.blueBias, params.greenBias, params.redBias]
    else:
        raise ValueError(
            "Unknown color space for tensor {}".format(operator.inputs[0].full_name)
        )

    if container.target_opset < 9:
        attrs = {"name": operator.full_name, "scale": params.channelScale}
        attrs["bias"] = bias
        container.add_node(
            "ImageScaler",
            [operator.inputs[0].full_name],
            [operator.outputs[0].full_name],
            **attrs
        )
    else:
        # In comments below, assume input tensor is X,
        # the scale scalar is a, the bias vector is b.

        # Store the scalar, a, used to scale all elements in the input tensor.
        aName = scope.get_unique_variable_name(operator.full_name + "_scale")
        container.add_initializer(
            aName, onnx_proto.TensorProto.FLOAT, [1], [params.channelScale]
        )

        # Store the bias vector. It will be added into the input tensor.
        bName = scope.get_unique_variable_name(operator.full_name + "_bias")
        container.add_initializer(
            bName, onnx_proto.TensorProto.FLOAT, [len(bias), 1, 1], bias
        )

        # Compute Z = a * X.
        zName = scope.get_unique_variable_name(operator.full_name + "_scaled")
        apply_mul(scope, [operator.input_full_names[0], aName], zName, container)

        # Compute Y = Z + b, which is the final output.
        apply_add(scope, [bName, zName], operator.output_full_names[0], container)


register_converter("scalerPreprocessor", convert_preprocessing_scaler)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_transpose
from ....common._registration import register_converter


def convert_permute(scope, operator, container):
    apply_transpose(
        scope,
        operator.input_full_names,
        operator.output_full_names,
        container,
        operator_name=operator.full_name,
        perm=operator.raw_operator.permute.axis,
    )


register_converter("permute", convert_permute)



# SPDX-License-Identifier: Apache-2.0

import copy
from onnx import onnx_pb as onnx_proto
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator


def convert_bucketizer(scope, operator, container):
    op = operator.raw_operator
    splits = op.getSplits()
    if splits[0] != -float("inf") or splits[-1] != float("inf"):
        raise RuntimeError("the Splits must include positive/negative infinity")

    input_shape = operator.inputs[0].type.shape
    reshape_data = scope.get_unique_variable_name("reshape_info_tensor")
    container.add_initializer(
        reshape_data, onnx_proto.TensorProto.INT64, [3], [-1, input_shape[1], 1]
    )
    outputs = []
    for split in splits[1:]:
        less_output = scope.get_unique_variable_name("less_output_tensor")
        initializer_name = "initializer_" + str(split)
        container.add_initializer(
            initializer_name, onnx_proto.TensorProto.FLOAT, [1], [split]
        )
        container.add_node(
            "Less",
            [operator.inputs[0].full_name, initializer_name],
            less_output,
            name=scope.get_unique_operator_name("Less"),
            op_version=7,
        )
        casted_output = scope.get_unique_variable_name("cast_output_tensor")
        container.add_node(
            "Cast",
            less_output,
            casted_output,
            name=scope.get_unique_operator_name("Cast"),
            op_version=6,
            to=1,
        )
        redim_output = scope.get_unique_variable_name("reshape_output_tensor")
        container.add_node(
            "Reshape",
            [casted_output, reshape_data],
            redim_output,
            name=scope.get_unique_operator_name("Reshape"),
            op_version=5,
        )
        outputs.append(redim_output)
    concat_output = scope.get_unique_variable_name("concat_output_tensor")
    container.add_node(
        "Concat",
        outputs,
        concat_output,
        name=scope.get_unique_operator_name("Concat"),
        op_version=1,
        axis=2,
    )
    argmax_output = scope.get_unique_variable_name("argmax_output_tensor")
    container.add_node(
        "ArgMax",
        concat_output,
        argmax_output,
        name=scope.get_unique_operator_name("ArgMax"),
        op_version=1,
        axis=2,
        keepdims=0,
    )
    container.add_node(
        "Cast",
        argmax_output,
        operator.output_full_names,
        name=scope.get_unique_operator_name("Cast"),
        op_version=6,
        to=1,
    )


register_converter("pyspark.ml.feature.Bucketizer", convert_bucketizer)


def calculate_bucketizer_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator,
        good_input_types=[FloatTensorType],
        good_output_types=[FloatTensorType],
    )
    input_type = copy.deepcopy(operator.inputs[0].type)
    for output in operator.outputs:
        output.type = input_type


register_shape_calculator(
    "pyspark.ml.feature.Bucketizer", calculate_bucketizer_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

import copy
import onnx
from ...common._registration import register_converter, register_shape_calculator
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common.data_types import FloatTensorType, Int64TensorType, StringTensorType


def convert_chi_sq_selector(scope, operator, container):
    op = operator.raw_operator
    indices = op.selectedFeatures
    indices_tensor = "indices_tensor"
    container.add_initializer(
        indices_tensor, onnx.TensorProto.INT64, [len(indices)], indices
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [operator.input_full_names[0], indices_tensor],
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )


register_converter("pyspark.ml.feature.ChiSqSelectorModel", convert_chi_sq_selector)


def calculate_chi_sq_selector_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType, StringTensorType]
    )
    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape[1] = len(operator.raw_operator.selectedFeatures)


register_shape_calculator(
    "pyspark.ml.feature.ChiSqSelectorModel", calculate_chi_sq_selector_shapes
)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from ...common._registration import register_converter, register_shape_calculator
from ...common.utils import check_input_and_output_numbers
from ...common.data_types import FloatTensorType


def convert_sparkml_vector_indexer(scope, operator, container):
    feature_count = operator.raw_operator.numFeatures
    category_map = operator.raw_operator.categoryMaps
    split_output_names = [
        scope.get_unique_variable_name("split_tensor_%d" % i)
        for i in range(0, feature_count)
    ]
    if feature_count > 1:
        container.add_node(
            "Split",
            operator.inputs[0].full_name,
            split_output_names,
            name=scope.get_unique_operator_name("Split"),
            op_version=2,
            axis=1,
            split=[1] * feature_count,
        )
    else:
        split_output_names = operator.input_full_names
    concat_inputs = split_output_names.copy()
    for i in category_map.keys():
        converted_output = scope.get_unique_variable_name("converted_tensor_%d" % i)
        container.add_node(
            "Cast",
            split_output_names[i],
            converted_output,
            name=scope.get_unique_operator_name("Cast"),
            op_version=9,
            to=onnx_proto.TensorProto.STRING,
        )
        attrs = {
            "name": scope.get_unique_operator_name("LabelEncoder"),
            "classes_strings": ["{0:g}".format(c) for c in category_map[i].keys()],
            "default_string": "__unknown__",
        }
        encoded_output_name = scope.get_unique_variable_name("indexed_tensor_%d" % i)
        container.add_node(
            "LabelEncoder",
            converted_output,
            encoded_output_name,
            op_domain="ai.onnx.ml",
            **attrs
        )
        converted_float_output = scope.get_unique_variable_name(
            "converted_float_tensor_%d" % i
        )
        if feature_count == 1:
            converted_float_output = operator.output_full_names[0]
        container.add_node(
            "Cast",
            encoded_output_name,
            converted_float_output,
            name=scope.get_unique_operator_name("Cast"),
            op_version=9,
            to=onnx_proto.TensorProto.FLOAT,
        )
        concat_inputs[i] = converted_float_output
    # add the final Concat
    if feature_count > 1:
        container.add_node(
            "Concat",
            concat_inputs,
            operator.output_full_names[0],
            name=scope.get_unique_operator_name("Concat"),
            op_version=4,
            axis=1,
        )


register_converter(
    "pyspark.ml.feature.VectorIndexerModel", convert_sparkml_vector_indexer
)


def calculate_vector_indexer_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = FloatTensorType([N, operator.raw_operator.numFeatures])


register_shape_calculator(
    "pyspark.ml.feature.VectorIndexerModel", calculate_vector_indexer_shapes
)



# SPDX-License-Identifier: Apache-2.0

import numbers
from ...common._apply_operation import apply_reshape
from ...common._registration import register_converter


def convert_tensor_to_probability_map(scope, operator, container):
    """
    This converter tries to convert a special operator
    'TensorToProbabilityMap' into a sequence of some ONNX operators.
    Those operators are used to create a dictionary in which keys
    are class labels and values are the associated
    probabilities. We assume that the elements in the given probability
    tensor are aligned with the class labels
    specified in the CoreML model.

    Notice that ONNX<1.2 doesn't support a CoreML classifier
    with a batch size larger than one because old ONNX ZipMap
    is not able to produce a sequence of dictionaries.
    This issue has been fixed in ONNX-1.2.
    """
    attrs = {"name": scope.get_unique_operator_name("ZipMap")}

    model_type = operator.raw_operator.WhichOneof("Type")
    if model_type == "neuralNetworkClassifier":
        model = operator.raw_operator.neuralNetworkClassifier
        if model.WhichOneof("ClassLabels") == "stringClassLabels":
            attrs["classlabels_strings"] = list(
                s.encode("utf-8") for s in model.stringClassLabels.vector
            )
        elif model.WhichOneof("ClassLabels") == "int64ClassLabels":
            attrs["classlabels_int64s"] = list(
                int(i) for i in model.int64ClassLabels.vector
            )
        else:
            raise ValueError("Unknown label type found")
    elif model_type == "pipelineClassifier":
        model = operator.raw_operator.pipelineClassifier
        if model.WhichOneof("ClassLabels") == "stringClassLabels":
            attrs["classlabels_strings"] = list(
                s.encode("utf-8") for s in model.stringClassLabels.vector
            )
        elif model.WhichOneof("ClassLabels") == "int64ClassLabels":
            attrs["classlabels_int64s"] = list(
                int(i) for i in model.int64ClassLabels.vector
            )
        else:
            raise ValueError("Unknown label type found")
    else:
        raise TypeError(
            "Only neural network classifiers and pipeline classifiers are supported"
        )

    input_shape = operator.inputs[0].type.shape
    if len(operator.inputs[0].type.shape) != 2:
        # Calculate the shape attribute of ONNX Reshape
        if input_shape[0] != "None":
            N = input_shape[0]
        else:
            # -1 means that this dimension is automatically
            # determined in runtime and unknown in conversion time
            N = -1

        if all(isinstance(i, numbers.Integral) for i in input_shape[1:]):
            C = 1
            for i in input_shape[1:]:
                C *= int(i)
        else:
            # -1 means that this dimension is automatically determined
            # in runtime and unknown in conversion time
            C = -1

        # ZipMap in ONNX only accepts [C] and [N, C] inputs.
        # In cases of [N, C, 1, 1], we reshape the probability tensor
        # into [N, C] before feeding it into ZipMap.
        buffer_name = scope.get_unique_variable_name("buffer")
        apply_reshape(
            scope,
            operator.inputs[0].full_name,
            buffer_name,
            container,
            desired_shape=[N, C],
        )
    else:
        buffer_name = operator.inputs[0].full_name

    container.add_node(
        "ZipMap",
        buffer_name,
        operator.outputs[0].full_name,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("tensorToProbabilityMap", convert_tensor_to_probability_map)



# SPDX-License-Identifier: Apache-2.0

from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import (
    apply_add,
    apply_mul,
    apply_div,
    apply_sub,
    apply_concat,
    apply_cast,
)
from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import FloatTensorType, DoubleTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from .tree_ensemble_common import save_read_sparkml_model_data

MinHashLSH_HASH_PRIME = 2038074743
g_rand_coefficients = None


def get_rand_coefficients(operator):
    global g_rand_coefficients
    if not g_rand_coefficients:
        g_rand_coefficients = save_read_sparkml_model_data(
            operator.raw_params["SparkSession"], operator.raw_operator
        ).first()[0]
    return g_rand_coefficients


def convert_min_hash_lsh(scope, operator, container):
    int_type = onnx_proto.TensorProto.INT64
    rand_coefficients = get_rand_coefficients(operator)
    coeffs = []
    for i in range(0, len(rand_coefficients), 2):
        coeffs.append((rand_coefficients[i], rand_coefficients[i + 1]))
    one = scope.get_unique_variable_name("one_tensor")
    container.add_initializer(one, int_type, [1], [1])
    prime = scope.get_unique_variable_name("prime_tensor")
    container.add_initializer(prime, int_type, [1], [MinHashLSH_HASH_PRIME])

    non_zeros_int = scope.get_unique_variable_name("non_zero_int_tensor")
    container.add_node(
        "NonZero", operator.input_full_names, non_zeros_int, op_version=9
    )
    non_zeros = scope.get_unique_variable_name("non_zeros_tensor")
    apply_cast(scope, non_zeros_int, non_zeros, container, to=int_type)
    remainders = []
    for coeff in coeffs:
        one_added = scope.get_unique_variable_name("one_added_tensor")
        apply_add(scope, [one, non_zeros], one_added, container)
        a = scope.get_unique_variable_name("a_coeff_tensor")
        container.add_initializer(a, int_type, [1], [coeff[0]])
        b = scope.get_unique_variable_name("b_coeff_tensor")
        container.add_initializer(b, int_type, [1], [coeff[1]])
        coeff_0_times = scope.get_unique_variable_name("coeff_0_times_tensor")
        apply_mul(scope, [a, one_added], coeff_0_times, container)
        coeff_1_added = scope.get_unique_variable_name("coeff_1_added_tensor")
        apply_add(scope, [b, coeff_0_times], coeff_1_added, container)
        # this is for Mod
        div_by_prime = scope.get_unique_variable_name("div_by_prime_tensor")
        apply_div(scope, [coeff_1_added, prime], div_by_prime, container)
        prime_x_floor = scope.get_unique_variable_name("prime_x_floor_tensor")
        apply_mul(scope, [div_by_prime, prime], prime_x_floor, container)
        remainder = scope.get_unique_variable_name("remainder_tensor")
        apply_sub(scope, [coeff_1_added, prime_x_floor], remainder, container)
        float_remainder = scope.get_unique_variable_name("float_remainder_tensor")
        apply_cast(
            scope,
            remainder,
            float_remainder,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        min_remainder = scope.get_unique_variable_name("min_remainder_tensor")
        container.add_node(
            "ReduceMin",
            float_remainder,
            min_remainder,
            op_version=1,
            axes=[1],
            keepdims=1,
        )
        remainders.append(min_remainder)
    apply_concat(scope, remainders, operator.output_full_names, container, axis=1)


register_converter("pyspark.ml.feature.MinHashLSHModel", convert_min_hash_lsh)


def calculate_min_hash_lsh_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, DoubleTensorType]
    )

    N = operator.inputs[0].type.shape[0]
    C = len(get_rand_coefficients(operator)) // 2
    operator.outputs[0].type = FloatTensorType([N, C])


register_shape_calculator(
    "pyspark.ml.feature.MinHashLSHModel", calculate_min_hash_lsh_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter, register_shape_calculator
from ....proto import onnx_proto
from ...common.data_types import FloatTensorType
from ...common._topology import Operator, Scope
from pyspark.ml.feature import OneHotEncoderModel
from typing import List


def _get_categories(operator: Operator) -> List[List[int]]:
    op: OneHotEncoderModel = operator.raw_operator
    categorySizes: List[int] = op.categorySizes

    # This is necessary to match SparkML's OneHotEncoder behavior.
    # If handleInvalid is set to "keep", an extra "category" indicating
    # invalid values is added as last category:
    #   - if dropLast is set to false, then an extra bit will be added for invalid input,
    #     which does not match ONNX's OneHotEncoder behavior.
    #   - if dropLast is set to true, then invalid values are encoded as all-zeros vector,
    #     which matches ONNX's OneHotEncoder behavior when "zeros" is set to 1.
    # If handleInvalid is set to "error", then:
    #   - if dropLast is set to false, then nothing will be added or dropped,
    #     and matches ONNX's OneHotEncoder behavior when "zeros" is set to 0.
    #   - if dropLast is set to true, then the last bit will be dropped,
    #     which does not match ONNX's OneHotEncoder behavior.
    if (op.getHandleInvalid() == "keep" and not op.getDropLast()) or (
        op.getHandleInvalid() == "error" and op.getDropLast()
    ):
        raise RuntimeError(
            f"The 'handleInvalid' and 'dropLast' parameters must be set to "
            f"('keep', True) or ('error', False), but got "
            f"('{op.getHandleInvalid()}', {op.getDropLast()}) instead."
        )

    return [list(range(0, size)) for size in categorySizes]


def convert_sparkml_one_hot_encoder(scope: Scope, operator: Operator, container):
    categories = _get_categories(operator)
    N = operator.inputs[0].type.shape[0] or -1

    zeros = 1 if operator.raw_operator.getHandleInvalid() == "keep" else 0

    for i in range(0, len(categories)):
        encoder_type = "OneHotEncoder"

        # Set zeros to 0 to match the "error"
        # handleInvalid behavior of SparkML's OneHotEncoder.
        encoder_attrs = {
            "name": scope.get_unique_operator_name(encoder_type),
            "cats_int64s": categories[i],
            "zeros": zeros,
        }

        encoded_feature_name = scope.get_unique_variable_name(
            "encoded_feature_at_" + str(i)
        )
        container.add_node(
            op_type=encoder_type,
            inputs=[operator.inputs[i].full_name],
            outputs=[encoded_feature_name],
            op_domain="ai.onnx.ml",
            op_version=1,
            **encoder_attrs,
        )

        shape_variable_name = scope.get_unique_variable_name("shape_at_" + str(i))
        container.add_initializer(
            shape_variable_name,
            onnx_proto.TensorProto.INT64,
            [2],
            [N, len(categories[i])],
        )

        container.add_node(
            op_type="Reshape",
            inputs=[encoded_feature_name, shape_variable_name],
            outputs=[operator.outputs[i].full_name],
            op_domain="ai.onnx",
            op_version=13,
        )


register_converter(
    "pyspark.ml.feature.OneHotEncoderModel", convert_sparkml_one_hot_encoder
)


def calculate_sparkml_one_hot_encoder_output_shapes(operator: Operator):
    categories = _get_categories(operator)
    N = operator.inputs[0].type.shape[0]
    for i, output in enumerate(operator.outputs):
        output.type = FloatTensorType([N, len(categories[i])])


register_shape_calculator(
    "pyspark.ml.feature.OneHotEncoderModel",
    calculate_sparkml_one_hot_encoder_output_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import warnings
import packaging.version as pv
import onnx
from .common import utils


def convert_coreml(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    if targeted_onnx is not None:
        warnings.warn(
            "targeted_onnx is deprecated. Use target_opset.", DeprecationWarning
        )
    if not utils.coreml_installed():
        raise RuntimeError(
            "coremltools is not installed. Please install coremltools to use this feature."
        )

    from .coreml.convert import convert

    return convert(
        model,
        name,
        initial_types,
        doc_string,
        target_opset,
        targeted_onnx,
        custom_conversion_functions,
        custom_shape_calculators,
    )


def convert_keras(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=None,
    channel_first_inputs=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
    default_batch_size=1,
):
    """
    .. versionchanged:: 1.9.0
        The conversion is now using *tf2onnx*.
    """
    if targeted_onnx is not None:
        warnings.warn(
            "targeted_onnx is deprecated and unused. Use target_opset.",
            DeprecationWarning,
        )
    import tensorflow as tf

    if pv.Version(tf.__version__) < pv.Version("2.0"):
        # Former converter for tensorflow<2.0.
        from keras2onnx import convert_keras as convert

        return convert(model, name, doc_string, target_opset, channel_first_inputs)
    else:
        # For tensorflow>=2.0, new converter based on tf2onnx.
        import tf2onnx

        if not utils.tf2onnx_installed():
            raise RuntimeError(
                "tf2onnx is not installed. Please install it to use this feature."
            )

        if custom_conversion_functions is not None:
            warnings.warn(
                "custom_conversion_functions is not supported any more. Please set it to None."
            )
        if custom_shape_calculators is not None:
            warnings.warn(
                "custom_shape_calculators is not supported any more. Please set it to None."
            )
        if default_batch_size != 1:
            warnings.warn(
                "default_batch_size is not supported any more. Please set it to 1."
            )
        if default_batch_size != 1:
            warnings.warn(
                "default_batch_size is not supported any more. Please set it to 1."
            )

        if initial_types is not None:
            from onnxconverter_common import (
                FloatTensorType,
                DoubleTensorType,
                Int64TensorType,
                Int32TensorType,
                StringTensorType,
                BooleanTensorType,
            )

            spec = []
            for name, kind in initial_types:
                if isinstance(kind, FloatTensorType):
                    dtype = tf.float32
                elif isinstance(kind, Int64TensorType):
                    dtype = tf.int64
                elif isinstance(kind, Int32TensorType):
                    dtype = tf.int32
                elif isinstance(kind, DoubleTensorType):
                    dtype = tf.float64
                elif isinstance(kind, StringTensorType):
                    dtype = tf.string
                elif isinstance(kind, BooleanTensorType):
                    dtype = tf.bool
                else:
                    raise TypeError(
                        "Unexpected type %r, cannot infer tensorflow type." % type(kind)
                    )
                spec.append(tf.TensorSpec(tuple(kind.shape), dtype, name=name))
            input_signature = tuple(spec)
        else:
            input_signature = None

        model_proto, external_tensor_storage = tf2onnx.convert.from_keras(
            model,
            input_signature=input_signature,
            opset=target_opset,
            custom_ops=None,
            custom_op_handlers=None,
            custom_rewriter=None,
            inputs_as_nchw=channel_first_inputs,
            extra_opset=None,
            shape_override=None,
            target=None,
            large_model=False,
            output_path=None,
        )
        if external_tensor_storage is not None:
            warnings.warn(
                "The current API does not expose the second result 'external_tensor_storage'. "
                "Use tf2onnx directly to get it."
            )
        model_proto.doc_string = doc_string
        return model_proto


def convert_libsvm(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    if targeted_onnx is not None:
        warnings.warn(
            "targeted_onnx is deprecated. Use target_opset.", DeprecationWarning
        )
    if not utils.libsvm_installed():
        raise RuntimeError(
            "libsvm is not installed. Please install libsvm to use this feature."
        )

    from .libsvm.convert import convert

    return convert(
        model,
        name,
        initial_types,
        doc_string,
        target_opset,
        targeted_onnx,
        custom_conversion_functions,
        custom_shape_calculators,
    )


def convert_catboost(
    model, name=None, initial_types=None, doc_string="", target_opset=None
):
    try:
        from catboost.utils import convert_to_onnx_object
    except ImportError:
        raise RuntimeError(
            "CatBoost is not installed or needs to be updated. "
            "Please install/upgrade CatBoost to use this feature."
        )

    return convert_to_onnx_object(
        model,
        export_parameters={"onnx_doc_string": doc_string, "onnx_graph_name": name},
        initial_types=initial_types,
        target_opset=target_opset,
    )


def convert_lightgbm(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
    without_onnx_ml=False,
    zipmap=True,
    split=None,
):
    if targeted_onnx is not None:
        warnings.warn(
            "targeted_onnx is deprecated. Use target_opset.", DeprecationWarning
        )
    if not utils.lightgbm_installed():
        raise RuntimeError(
            "lightgbm is not installed. Please install lightgbm to use this feature."
        )

    from .lightgbm.convert import convert

    return convert(
        model,
        name,
        initial_types,
        doc_string,
        target_opset,
        targeted_onnx,
        custom_conversion_functions,
        custom_shape_calculators,
        without_onnx_ml,
        zipmap=zipmap,
        split=split,
    )


def convert_sklearn(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    if targeted_onnx is not None:
        warnings.warn(
            "targeted_onnx is deprecated. Use target_opset.", DeprecationWarning
        )
    if not utils.sklearn_installed():
        raise RuntimeError(
            "scikit-learn is not installed. Please install scikit-learn to use this feature."
        )

    if not utils.skl2onnx_installed():
        raise RuntimeError(
            "skl2onnx is not installed. Please install skl2onnx to use this feature."
        )

    from skl2onnx.convert import convert_sklearn as convert_skl2onnx

    return convert_skl2onnx(
        model,
        name,
        initial_types,
        doc_string,
        target_opset,
        custom_conversion_functions,
        custom_shape_calculators,
    )


def convert_sparkml(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
    spark_session=None,
):
    if targeted_onnx is not None:
        warnings.warn(
            "targeted_onnx is deprecated. Use target_opset.", DeprecationWarning
        )
    if not utils.sparkml_installed():
        raise RuntimeError(
            "Spark is not installed. Please install Spark to use this feature."
        )

    from .sparkml.convert import convert

    return convert(
        model,
        name,
        initial_types,
        doc_string,
        target_opset,
        targeted_onnx,
        custom_conversion_functions,
        custom_shape_calculators,
        spark_session,
    )


def convert_xgboost(*args, **kwargs):
    if kwargs.get("targeted_onnx", None) is not None:
        warnings.warn(
            "targeted_onnx is deprecated. Use target_opset.", DeprecationWarning
        )
    if not utils.xgboost_installed():
        raise RuntimeError(
            "xgboost is not installed. Please install xgboost to use this feature."
        )

    from .xgboost.convert import convert

    return convert(*args, **kwargs)


def convert_h2o(*args, **kwargs):
    if kwargs.get("targeted_onnx", None) is not None:
        warnings.warn(
            "targeted_onnx is deprecated. Use target_opset.", DeprecationWarning
        )
    if not utils.h2o_installed():
        raise RuntimeError(
            "h2o is not installed. Please install h2o to use this feature."
        )

    from .h2o.convert import convert

    return convert(*args, **kwargs)


def _collect_input_nodes(graph, outputs):
    nodes_to_keep = set()
    input_nodes = set()
    node_inputs = [graph.get_tensor_by_name(ts_).op for ts_ in outputs]
    while node_inputs:
        nd_ = node_inputs[0]
        del node_inputs[0]
        if nd_.type in ["Placeholder", "PlaceholderV2", "PlaceholderWithDefault"]:
            input_nodes.add(nd_)
        if nd_ in nodes_to_keep:
            continue

        nodes_to_keep.add(nd_)
        node_inputs.extend(in_.op for in_ in nd_.inputs)

    return input_nodes, nodes_to_keep


def _convert_tf_wrapper(
    frozen_graph_def,
    name=None,
    input_names=None,
    output_names=None,
    doc_string="",
    target_opset=None,
    channel_first_inputs=None,
    debug_mode=False,
    custom_op_conversions=None,
    **kwargs
):
    """
    convert a tensorflow graph def into a ONNX model proto, just like how keras does.
    :param graph_def: the frozen tensorflow graph
    :param name: the converted onnx model internal name
    :param input_names: the inputs name list of the model
    :param output_names: the output name list of the model
    :param doc_string: doc string
    :param target_opset: the targeted onnx model opset
    :param channel_first_inputs: A list of channel first input (not supported yet)
    :param debug_mode: will enable the log and try to convert as much as possible on conversion
    :param kwargs: additional parameters of function `processs_tf_graph
        <https://github.com/onnx/tensorflow-onnx#creating-custom-op-mappings-from-python>`_
    :return an ONNX ModelProto
    """
    import tensorflow as tf
    import tf2onnx

    if target_opset is None:
        target_opset = onnx.defs.onnx_opset_version()

    if not doc_string:
        doc_string = "converted from {}".format(name)

    tf_graph_def = tf2onnx.tfonnx.tf_optimize(
        input_names, output_names, frozen_graph_def, True
    )
    with tf.Graph().as_default() as tf_graph:
        tf.import_graph_def(tf_graph_def, name="")

        if not input_names:
            input_nodes = list(_collect_input_nodes(tf_graph, output_names)[0])
            input_names = [nd_.outputs[0].name for nd_ in input_nodes]
        g = tf2onnx.tfonnx.process_tf_graph(
            tf_graph,
            continue_on_error=debug_mode,
            opset=target_opset,
            custom_op_handlers=custom_op_conversions,
            inputs_as_nchw=channel_first_inputs,
            output_names=output_names,
            input_names=input_names,
            **kwargs
        )

        onnx_graph = tf2onnx.optimizer.optimize_graph(g)
        model_proto = onnx_graph.make_model(doc_string)

    return model_proto


def convert_tensorflow(
    frozen_graph_def,
    name=None,
    input_names=None,
    output_names=None,
    doc_string="",
    target_opset=None,
    channel_first_inputs=None,
    debug_mode=False,
    custom_op_conversions=None,
    **kwargs
):
    import pkgutil

    if not pkgutil.find_loader("tf2onnx"):
        raise RuntimeError(
            "tf2onnx is not installed, please install it before calling this function."
        )

    return _convert_tf_wrapper(
        frozen_graph_def,
        name,
        input_names,
        output_names,
        doc_string,
        target_opset,
        channel_first_inputs,
        debug_mode,
        custom_op_conversions,
        **kwargs
    )



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_split
from ....common._registration import register_converter


def convert_split(scope, operator, container):
    # ONNX Split may evenly divide the input along the specified
    # axis if "split" attribute is not specified.
    # Also, CoreML always evenly split the input along C-axis.
    # Consequently, we only need to specify the axis
    # and make sure the number of outputs in ONNX matches that in CoreML.
    apply_split(
        scope,
        operator.input_full_names,
        operator.output_full_names,
        container,
        operator_name=operator.full_name,
        axis=1,
    )


register_converter("split", convert_split)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ...common._registration import register_converter


def convert_glm_regressor(scope, operator, container):
    from coremltools.proto.GLMRegressor_pb2 import GLMRegressor

    op_type = "LinearRegressor"
    glm = operator.raw_operator.glmRegressor
    attrs = {"name": operator.full_name}

    transform_table = {
        GLMRegressor.NoTransform: "NONE",
        GLMRegressor.Logit: "LOGISTIC",
        GLMRegressor.Probit: "PROBIT",
    }
    if glm.postEvaluationTransform in transform_table:
        attrs["post_transform"] = transform_table[glm.postEvaluationTransform]
    else:
        raise ValueError(
            "Unsupported post-transformation: {}".format(glm.postEvaluationTransform)
        )

    # Determine the dimensionality of the model weights. Conceptually,
    # the shape of the weight matrix in CoreML is E-by-F, where E and F
    # respectively denote the number of target variables and the number
    # of used features. Note that in ONNX, the shape is F-by-E.
    dim_target = len(glm.weights)
    dim_feature = len(glm.weights[0].value)

    matrix_w = np.ndarray(shape=(dim_feature, dim_target))
    for i, w in enumerate(glm.weights):
        matrix_w[:, i] = w.value

    attrs["targets"] = dim_target
    attrs["coefficients"] = matrix_w.flatten()
    attrs["intercepts"] = glm.offset

    container.add_node(
        op_type,
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("glmRegressor", convert_glm_regressor)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_padding_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N, C, H', W']
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)

    params = operator.raw_operator.padding
    if len(params.paddingAmounts.borderAmounts) > 0:
        output_shape[2] += params.paddingAmounts.borderAmounts[0].startEdgeSize
        output_shape[2] += params.paddingAmounts.borderAmounts[0].endEdgeSize
        output_shape[3] += params.paddingAmounts.borderAmounts[1].startEdgeSize
        output_shape[3] += params.paddingAmounts.borderAmounts[1].endEdgeSize

    operator.outputs[0].type.shape = output_shape


register_shape_calculator("padding", calculate_padding_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_inner_product_output_shapes(operator):
    """
    Allowed input/output patterns are
        1. [N, C] ---> [N, C']
        2. [N, C, 1, 1] ---> [N, C', 1, 1]
    """
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    input = operator.inputs[0]
    output = operator.outputs[0]

    input_shape = input.type.shape
    if len(input_shape) == 4 and (input_shape[2] != 1 or input_shape[3] != 1):
        raise RuntimeError("If input is a 4-D tensor, its shape must be [N, C, 1, 1]")

    params = operator.raw_operator.innerProduct

    if input_shape[1] != params.inputChannels:
        raise RuntimeError(
            "Dimension mismatch along C-axis. Expected %s but got %s"
            % (params.inputChannels, input_shape[1])
        )

    if len(input_shape) == 4:
        output.type.shape = [input_shape[0], params.outputChannels, 1, 1]
    elif len(input_shape) == 2:
        output.type.shape = [input_shape[0], params.outputChannels]
    else:
        raise RuntimeError("Input must be a 2-D or a 4-D tensor")


register_shape_calculator("innerProduct", calculate_inner_product_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import onnx
from ....common._apply_operation import (
    apply_abs,
    apply_log,
    apply_exp,
    apply_reciprocal,
    apply_sqrt,
    apply_clip,
    apply_pow,
    apply_affine,
)
from ....common._registration import register_converter


def convert_unary(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import UnaryFunctionLayerParams as Params

    params = operator.raw_operator.unary
    preprocessor_name = scope.get_unique_operator_name("Affine")
    preprocessed_variable_name = scope.get_unique_variable_name(
        preprocessor_name + "_output"
    )
    apply_affine(
        scope,
        operator.input_full_names[0],
        preprocessed_variable_name,
        container,
        operator_name=preprocessor_name,
        alpha=params.scale,
        beta=params.shift,
    )

    if params.type == Params.RSQRT:
        sqrt_tensor_name = scope.get_unique_variable_name(
            operator.full_name + "_intra_tensor"
        )
        apply_sqrt(scope, preprocessed_variable_name, sqrt_tensor_name, container)
        apply_reciprocal(scope, sqrt_tensor_name, operator.output_full_names, container)
    elif params.type == Params.POWER:
        exp_name = scope.get_unique_variable_name("Y")
        container.add_initializer(exp_name, onnx.TensorProto.FLOAT, [], [params.alpha])

        apply_pow(
            scope,
            [preprocessed_variable_name, exp_name],
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
            broadcast=1,
        )
    elif params.type == Params.THRESHOLD:
        apply_clip(
            scope,
            preprocessed_variable_name,
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
            min=params.alpha,
        )
    elif params.type == Params.SQRT:
        apply_sqrt(
            scope,
            preprocessed_variable_name,
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
        )
    elif params.type == Params.INVERSE:
        apply_reciprocal(
            scope,
            preprocessed_variable_name,
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
        )
    elif params.type == Params.EXP:
        apply_exp(
            scope,
            preprocessed_variable_name,
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
        )
    elif params.type == Params.LOG:
        apply_log(
            scope,
            preprocessed_variable_name,
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
        )
    elif params.type == Params.ABS:
        apply_abs(
            scope,
            preprocessed_variable_name,
            operator.output_full_names,
            container,
            operator_name=operator.full_name,
        )
    else:
        raise ValueError("Unsupported unary function :{}".format(params.type))


register_converter("unary", convert_unary)



# SPDX-License-Identifier: Apache-2.0

import copy
from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_concat, apply_cast
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_polynomial_expansion(scope, operator, container):
    op = operator.raw_operator
    feature_count = operator.inputs[0].type.shape[1]
    degree = op.getDegree()
    all_combinations = calc_combinations(feature_count, degree)
    transformed_columns = []

    for i, comb in enumerate(all_combinations):
        if comb is None:
            pass
        else:
            comb_name = scope.get_unique_variable_name("comb")
            col_name = scope.get_unique_variable_name("col")
            prod_name = scope.get_unique_variable_name("prod")

            container.add_initializer(
                comb_name, onnx_proto.TensorProto.INT64, [len(comb)], list(comb)
            )

            container.add_node(
                "ArrayFeatureExtractor",
                [operator.inputs[0].full_name, comb_name],
                col_name,
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
                op_domain="ai.onnx.ml",
            )
            reduce_prod_input = col_name
            if (
                operator.inputs[0].type._get_element_onnx_type()
                == onnx_proto.TensorProto.INT64
            ):
                float_col_name = scope.get_unique_variable_name("col")
                container.add_node(
                    "Cast",
                    col_name,
                    float_col_name,
                    name=scope.get_unique_operator_name("Cast"),
                    to=onnx_proto.TensorProto.FLOAT,
                )
                reduce_prod_input = float_col_name

            container.add_node(
                "ReduceProd",
                reduce_prod_input,
                prod_name,
                axes=[1],
                name=scope.get_unique_operator_name("ReduceProd"),
            )
            transformed_columns.append(prod_name)

    if operator.inputs[0].type._get_element_onnx_type() == onnx_proto.TensorProto.INT64:
        concat_result_name = scope.get_unique_variable_name("concat_result")

        apply_concat(
            scope,
            [t for t in transformed_columns],
            concat_result_name,
            container,
            axis=1,
        )
        apply_cast(
            scope,
            concat_result_name,
            operator.outputs[0].full_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:
        apply_concat(
            scope,
            [t for t in transformed_columns],
            operator.outputs[0].full_name,
            container,
            axis=1,
        )


register_converter(
    "pyspark.ml.feature.PolynomialExpansion", convert_sparkml_polynomial_expansion
)


def expand_inner(values, last_index, degree, multiplier, poly_values, cur_poly_index):
    if degree == 0 or last_index < 0:
        if cur_poly_index >= 0:
            # skip the very first 1
            poly_values[cur_poly_index] = multiplier.copy()
    else:
        v = values[last_index]
        last_index1 = last_index - 1
        alpha = multiplier.copy()
        i = 0
        cur_start = cur_poly_index
        while i <= degree:
            cur_start, multiplier = expand_inner(
                values, last_index1, degree - i, alpha, poly_values, cur_start
            )
            i += 1
            alpha.append(v)
    return cur_poly_index + get_poly_size(last_index + 1, degree), multiplier


def calc_combinations(feature_count, degree):
    values = range(0, feature_count)
    poly_size = get_poly_size(feature_count, degree)
    poly_values = [None] * poly_size
    expand_inner(values, feature_count - 1, degree, [], poly_values, -1)
    return poly_values


def get_poly_size(feature_count, degree):
    return get_combinations_count(feature_count + degree, degree)


def get_combinations_count(n, k):
    from math import factorial

    if n == k or k == 0:
        return 1
    if k == 1 or k == n - 1:
        return n
    return factorial(n) // factorial(k) // factorial(n - k)


def calculate_sparkml_polynomial_expansion_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(
        operator, good_input_types=[FloatTensorType, Int64TensorType]
    )

    N = operator.inputs[0].type.shape[0]
    C = get_combinations_count(
        operator.inputs[0].type.shape[1], operator.raw_operator.getDegree()
    )
    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape = [N, C]


register_shape_calculator(
    "pyspark.ml.feature.PolynomialExpansion",
    calculate_sparkml_polynomial_expansion_output_shapes,
)
